__all__ = [
    "get_bolt_parsers",
    "get_stream_lexical_scope",
    "get_stream_macro_scope",
    "get_stream_pending_macros",
    "ToplevelHandler",
    "create_bolt_root_parser",
    "create_bolt_command_parser",
    "UndefinedIdentifier",
    "ImprovedErrorHandler",
    "parse_statement",
    "BranchScopeManager",
    "FinalExpressionParser",
    "InterpolationParser",
    "DisableInterpolationParser",
    "ForkLexicalScopeParser",
    "AssignmentStatementParser",
    "EscapeAnalysisParser",
    "EscapeAnalysisResolver",
    "MemoRootParser",
    "MemoHandler",
    "parse_decorator",
    "AssignmentTargetParser",
    "IfElseLoweringParser",
    "VanillaReturnHandler",
    "BreakContinueConstraint",
    "parse_deferred_root",
    "parse_function_signature",
    "parse_proc_macro_signature",
    "MacroMatchParser",
    "MacroHandler",
    "ProcMacroParser",
    "ProcMacroExpansion",
    "parse_class_name",
    "ClassBasesParser",
    "parse_class_root",
    "parse_del_target",
    "parse_identifier",
    "TrailingCommaParser",
    "ImportLocationConstraint",
    "ImportStatementHandler",
    "parse_python_import",
    "parse_import_name",
    "parse_name_list",
    "BindingStorageHandler",
    "DocstringHandler",
    "WithContextParser",
    "FlushPendingBindingsParser",
    "SubcommandConstraint",
    "DeferredRootBacktracker",
    "LexicalScopeConstraint",
    "RootScopeHandler",
    "BinaryParser",
    "UnaryParser",
    "ChainedComparisonParser",
    "UnpackParser",
    "UnpackConstraint",
    "KeywordParser",
    "LookupParser",
    "PrimaryParser",
    "BuiltinCallRestriction",
    "parse_dict_item",
    "LiteralParser",
]


import re
from dataclasses import dataclass, field, replace
from typing import Any, Dict, FrozenSet, List, Literal, Optional, Set, Tuple, Type, cast
from uuid import UUID, uuid4

from beet import Function
from beet.core.utils import extra_field
from mecha import (
    AdjacentConstraint,
    AlternativeParser,
    AstChildren,
    AstCommand,
    AstJson,
    AstNode,
    AstNumber,
    AstResourceLocation,
    AstRoot,
    BasicLiteralParser,
    CommandSpec,
    CommandTree,
    CommentDisambiguation,
    CompilationDatabase,
    MultilineParser,
    Parser,
    consume_line_continuation,
    delegate,
    get_stream_properties,
    get_stream_scope,
    get_stream_spec,
)
from mecha.contrib.relative_location import resolve_using_database
from mecha.utils import (
    JsonQuoteHelper,
    QuoteHelper,
    normalize_whitespace,
    string_to_number,
)
from tokenstream import (
    InvalidSyntax,
    Token,
    TokenStream,
    UnexpectedEOF,
    UnexpectedToken,
    set_location,
)

from .ast import (
    AstAssignment,
    AstAttribute,
    AstCall,
    AstChainedComparison,
    AstClassBases,
    AstClassName,
    AstClassRoot,
    AstDecorator,
    AstDeferredRoot,
    AstDict,
    AstDictItem,
    AstDictUnquotedKey,
    AstDocstring,
    AstEscapeAnalysisRoot,
    AstEscapeRoot,
    AstExpression,
    AstExpressionBinary,
    AstExpressionUnary,
    AstFormatString,
    AstFormattedLocation,
    AstFromImport,
    AstFunctionSignature,
    AstFunctionSignatureArgument,
    AstFunctionSignatureElement,
    AstFunctionSignaturePositionalMarker,
    AstFunctionSignatureVariadicArgument,
    AstFunctionSignatureVariadicKeywordArgument,
    AstFunctionSignatureVariadicMarker,
    AstIdentifier,
    AstImportedItem,
    AstImportedMacro,
    AstInterpolation,
    AstKeyword,
    AstList,
    AstLookup,
    AstMacro,
    AstMacroArgument,
    AstMacroCall,
    AstMacroLiteral,
    AstMacroMatch,
    AstMacroMatchArgument,
    AstMacroMatchLiteral,
    AstMemo,
    AstNonFunctionRoot,
    AstProcMacro,
    AstProcMacroMarker,
    AstProcMacroResult,
    AstSlice,
    AstStatement,
    AstTarget,
    AstTargetAttribute,
    AstTargetIdentifier,
    AstTargetItem,
    AstTargetUnpack,
    AstTuple,
    AstTypeDeclaration,
    AstUnpack,
    AstValue,
    AstWithClause,
    AstWithContext,
)
from .emit import CommandEmitter
from .module import ModuleManager, UnusableCompilationUnit
from .pattern import (
    DOCSTRING_PATTERN,
    FALSE_PATTERN,
    IDENTIFIER_PATTERN,
    MODULE_PATTERN,
    NULL_PATTERN,
    NUMBER_PATTERN,
    RESOURCE_LOCATION_PATTERN,
    STRING_PATTERN,
    TRUE_PATTERN,
)
from .semantics import (
    Binding,
    ClassScope,
    FunctionScope,
    GlobalScope,
    LexicalScope,
    MacroScope,
    ProcMacroScope,
    UndefinedIdentifier,
)
from .utils import internal

IMPORT_REGEX = re.compile(rf"^{MODULE_PATTERN}$")


def get_bolt_parsers(
    parsers: Dict[str, Parser],
    modules: ModuleManager,
    bolt_prototypes: Set[str],
) -> Dict[str, Parser]:
    """Return the bolt parsers."""
    macro_handler = MacroHandler(
        parser=parsers["command"],
        proc_macro_overloads=(
            modules.cache.json.setdefault("proc_macro_overloads", {})
            if modules.cache
            else {}
        ),
    )

    return {
        ################################################################################
        # Command
        ################################################################################
        "root": ToplevelHandler(
            parser=create_bolt_root_parser(parsers["root"], macro_handler),
            modules=modules,
            macro_handler=macro_handler,
        ),
        "nested_root": create_bolt_root_parser(parsers["nested_root"], macro_handler),
        "root_item": ImprovedErrorHandler(
            AlternativeParser([parsers["root_item"], parse_statement])
        ),
        "command": create_bolt_command_parser(macro_handler, modules, bolt_prototypes),
        "command:argument:bolt:if_block": delegate("bolt:if_block"),
        "command:argument:bolt:elif_condition": delegate("bolt:elif_condition"),
        "command:argument:bolt:elif_block": delegate("bolt:elif_block"),
        "command:argument:bolt:else_block": delegate("bolt:else_block"),
        "command:argument:bolt:assignment_target": delegate("bolt:assignment_target"),
        "command:argument:bolt:with_context": delegate("bolt:with_context"),
        "command:argument:bolt:import": delegate("bolt:import"),
        "command:argument:bolt:import_name": delegate("bolt:import_name"),
        "command:argument:bolt:global_name": delegate("bolt:global_name"),
        "command:argument:bolt:nonlocal_name": delegate("bolt:nonlocal_name"),
        "command:argument:bolt:expression": delegate("bolt:expression"),
        "command:argument:bolt:deferred_root": delegate("bolt:deferred_root"),
        "command:argument:bolt:function_signature": delegate("bolt:function_signature"),
        "command:argument:bolt:macro_name": delegate("bolt:macro_name"),
        "command:argument:bolt:macro_signature": delegate("bolt:macro_signature"),
        "command:argument:bolt:macro_match": delegate("bolt:macro_match"),
        "command:argument:bolt:proc_macro": delegate("bolt:proc_macro"),
        "command:argument:bolt:class_name": delegate("bolt:class_name"),
        "command:argument:bolt:class_bases": delegate("bolt:class_bases"),
        "command:argument:bolt:class_root": delegate("bolt:class_root"),
        "command:argument:bolt:memo_variable": delegate("bolt:memo_variable"),
        "command:argument:bolt:memo_root": delegate("bolt:memo_root"),
        "command:argument:bolt:del_target": delegate("bolt:del_target"),
        ################################################################################
        # Bolt
        ################################################################################
        "bolt:if_block": BranchScopeManager(
            parser=delegate("nested_root"),
            init=True,
            fork=True,
        ),
        "bolt:elif_condition": BranchScopeManager(delegate("bolt:expression")),
        "bolt:elif_block": BranchScopeManager(
            parser=delegate("nested_root"),
            fork=True,
        ),
        "bolt:else_block": BranchScopeManager(
            parser=delegate("nested_root"),
            fork=True,
        ),
        "bolt:statement": AlternativeParser(
            [
                ForkLexicalScopeParser(
                    FlushPendingBindingsParser(
                        FinalExpressionParser(
                            AssignmentStatementParser(
                                delegate("bolt:augmented_assignment_target")
                            )
                        ),
                        after=True,
                    )
                ),
                ForkLexicalScopeParser(
                    FlushPendingBindingsParser(
                        FinalExpressionParser(
                            AssignmentStatementParser(
                                delegate("bolt:assignment_target"),
                                binding_only=True,
                            )
                        ),
                        after=True,
                    )
                ),
                FinalExpressionParser(
                    AssignmentStatementParser(delegate("bolt:expression"))
                ),
                FinalExpressionParser(delegate("bolt:decorator")),
            ]
        ),
        "bolt:decorator": parse_decorator,
        "bolt:assignment_target": AssignmentTargetParser(),
        "bolt:augmented_assignment_target": AssignmentTargetParser(
            require_local_binding=True, require_single=True
        ),
        "bolt:deferred_root": parse_deferred_root,
        "bolt:function_signature": parse_function_signature,
        "bolt_macro_literal": delegate("bolt:macro_literal"),
        "bolt:macro_literal": BasicLiteralParser(AstMacroLiteral),
        "bolt_macro_argument": delegate("bolt:macro_argument"),
        "bolt:macro_argument": BasicLiteralParser(AstMacroArgument),
        "bolt:macro_name": delegate("bolt:macro_literal"),
        "bolt:macro_signature": parse_proc_macro_signature,
        "bolt:macro_match": MacroMatchParser(
            literal_parser=delegate("bolt:macro_literal"),
            argument_parser=delegate("bolt:macro_argument"),
            resource_location_parser=DisableInterpolationParser(
                delegate("resource_location")
            ),
            json_properties_parser=DisableInterpolationParser(
                AdjacentConstraint(MultilineParser(delegate("json_object")), r"\{")
            ),
        ),
        "bolt:proc_macro": ProcMacroParser(modules),
        "bolt:class_name": parse_class_name,
        "bolt:class_bases": ClassBasesParser(
            AlternativeParser(
                [
                    KeywordParser(delegate("bolt:expression")),
                    delegate("bolt:expression"),
                ]
            )
        ),
        "bolt:class_root": FlushPendingBindingsParser(parse_class_root, after=True),
        "bolt:memo_variable": TrailingCommaParser(
            AlternativeParser(
                [
                    ForkLexicalScopeParser(
                        FlushPendingBindingsParser(
                            AssignmentStatementParser(
                                AssignmentTargetParser(require_single=True),
                                binding_only=True,
                            ),
                            after=True,
                        )
                    ),
                    delegate("bolt:interpolation"),
                ]
            )
        ),
        "bolt:memo_root": MemoRootParser(EscapeAnalysisParser(delegate("nested_root"))),
        "bolt:del_target": parse_del_target,
        "bolt:interpolation": BuiltinCallRestriction(
            PrimaryParser(delegate("bolt:identifier"), truncate=True),
            builtins=modules.builtins,
        ),
        "bolt:identifier": parse_identifier,
        "bolt:with_context": WithContextParser(
            expression_parser=delegate("bolt:expression"),
            target_parser=FlushPendingBindingsParser(
                AssignmentTargetParser(require_single=True),
                after=True,
            ),
        ),
        "bolt:import": AlternativeParser(
            [
                ImportLocationConstraint(
                    DisableInterpolationParser(delegate("resource_location_or_tag"))
                ),
                parse_python_import,
            ]
        ),
        "bolt:import_name": TrailingCommaParser(parse_import_name),
        "bolt:global_name": parse_name_list,
        "bolt:nonlocal_name": parse_name_list,
        "bolt:type_annotation": delegate("bolt:expression"),
        "bolt:expression": delegate("bolt:disjunction"),
        "bolt:disjunction": BinaryParser(
            operators=[r"\bor\b"],
            parser=delegate("bolt:conjunction"),
        ),
        "bolt:conjunction": BinaryParser(
            operators=[r"\band\b"],
            parser=delegate("bolt:inversion"),
        ),
        "bolt:inversion": UnaryParser(
            operators=[r"\bnot\b"],
            parser=delegate("bolt:comparison"),
        ),
        "bolt:comparison": ChainedComparisonParser(
            operators=[
                "==",
                "!=",
                "<=",
                "<",
                ">=",
                ">",
                r"\bnot\s+in\b",
                r"\bin\b",
                r"\bis\s+not\b",
                r"\bis\b",
            ],
            parser=delegate("bolt:bitwise_or"),
        ),
        "bolt:bitwise_or": BinaryParser(
            operators=[r"\|(?!=)"],
            parser=delegate("bolt:bitwise_xor"),
        ),
        "bolt:bitwise_xor": BinaryParser(
            operators=[r"\^(?!=)"],
            parser=delegate("bolt:bitwise_and"),
        ),
        "bolt:bitwise_and": BinaryParser(
            operators=[r"&(?!=)"],
            parser=delegate("bolt:shift_expr"),
        ),
        "bolt:shift_expr": BinaryParser(
            operators=[r"<<(?!=)", r">>(?!=)"],
            parser=delegate("bolt:sum"),
        ),
        "bolt:sum": BinaryParser(
            operators=[r"\+(?!=)", r"-(?!=)"],
            parser=delegate("bolt:term"),
        ),
        "bolt:term": BinaryParser(
            operators=[r"\*(?!=)", r"//(?!=)", r"/(?!=)", r"%(?!=)"],
            parser=delegate("bolt:factor"),
        ),
        "bolt:factor": UnaryParser(
            operators=[r"\+", "-"],
            parser=delegate("bolt:power"),
        ),
        "bolt:power": BinaryParser(
            operators=[r"\*\*(?!=)"],
            parser=delegate("bolt:primary"),
            right_associative=True,
        ),
        "bolt:lookup_argument": LookupParser(delegate("bolt:expression")),
        "bolt:call_argument": AlternativeParser(
            [
                KeywordParser(delegate("bolt:expression")),
                UnpackParser(delegate("bolt:expression")),
                delegate("bolt:expression"),
            ]
        ),
        "bolt:primary": PrimaryParser(delegate("bolt:atom")),
        "bolt:atom": AlternativeParser(
            [
                delegate("bolt:identifier"),
                delegate("bolt:literal"),
            ]
        ),
        "bolt:list_item": AlternativeParser(
            [
                UnpackConstraint(
                    type="list",
                    parser=UnpackParser(delegate("bolt:expression")),
                ),
                delegate("bolt:expression"),
            ]
        ),
        "bolt:dict_item": AlternativeParser(
            [
                UnpackConstraint(
                    type="dict",
                    parser=UnpackParser(delegate("bolt:expression")),
                ),
                parse_dict_item,
            ]
        ),
        "bolt:literal": LiteralParser(modules.database),
        ################################################################################
        # Interpolation
        ################################################################################
        "bool": AlternativeParser([parsers["bool"], InterpolationParser("bool")]),
        "numeric": AlternativeParser(
            [parsers["numeric"], InterpolationParser("numeric", prefix="-")]
        ),
        "coordinate": AlternativeParser(
            [
                parsers["coordinate"],
                InterpolationParser("coordinate", prefix="[~^]-?|-"),
            ]
        ),
        "time": AlternativeParser([parsers["time"], InterpolationParser("time")]),
        "word": AlternativeParser([InterpolationParser("word"), parsers["word"]]),
        "phrase": AlternativeParser([InterpolationParser("phrase"), parsers["phrase"]]),
        "greedy": AlternativeParser([InterpolationParser("greedy"), parsers["greedy"]]),
        "json": AlternativeParser([InterpolationParser("json"), parsers["json"]]),
        "json_object_entry": AlternativeParser(
            [InterpolationParser("json", unpack=r"\*\*"), parsers["json_object_entry"]]
        ),
        "json_array_element": AlternativeParser(
            [InterpolationParser("json", unpack=r"\*"), parsers["json_array_element"]]
        ),
        "json_object": AlternativeParser(
            [InterpolationParser("json_object"), parsers["json_object"]]
        ),
        "nbt": AlternativeParser([InterpolationParser("nbt"), parsers["nbt"]]),
        "nbt_compound_entry": AlternativeParser(
            [InterpolationParser("nbt", unpack=r"\*\*"), parsers["nbt_compound_entry"]]
        ),
        "nbt_list_or_array_element": AlternativeParser(
            [
                InterpolationParser("nbt", unpack=r"\*"),
                parsers["nbt_list_or_array_element"],
            ]
        ),
        "nbt_compound": AlternativeParser(
            [InterpolationParser("nbt_compound"), parsers["nbt_compound"]]
        ),
        "nbt_path": AlternativeParser(
            [InterpolationParser("nbt_path"), parsers["nbt_path"]]
        ),
        "range": AlternativeParser([parsers["range"], InterpolationParser("range")]),
        "resource_location_or_tag": CommentDisambiguation(
            AlternativeParser(
                [
                    InterpolationParser("resource_location", prefix=r"#"),
                    parsers["resource_location_or_tag"],
                ]
            )
        ),
        "item_slot": AlternativeParser(
            [parsers["item_slot"], InterpolationParser("item_slot")]
        ),
        "item_slots": AlternativeParser(
            [parsers["item_slots"], InterpolationParser("item_slots")]
        ),
        "uuid": AlternativeParser([InterpolationParser("uuid"), parsers["uuid"]]),
        "objective": AlternativeParser(
            [InterpolationParser("objective"), parsers["objective"]]
        ),
        "objective_criteria": AlternativeParser(
            [InterpolationParser("objective_criteria"), parsers["objective_criteria"]]
        ),
        "scoreboard_slot": AlternativeParser(
            [parsers["scoreboard_slot"], InterpolationParser("scoreboard_slot")]
        ),
        "swizzle": AlternativeParser(
            [parsers["swizzle"], InterpolationParser("swizzle")]
        ),
        "team": AlternativeParser([InterpolationParser("team"), parsers["team"]]),
        "advancement_predicate": AlternativeParser(
            [
                InterpolationParser("advancement_predicate"),
                parsers["advancement_predicate"],
            ]
        ),
        "color": AlternativeParser([InterpolationParser("color"), parsers["color"]]),
        "sort_order": AlternativeParser(
            [parsers["sort_order"], InterpolationParser("sort_order")]
        ),
        "gamemode": AlternativeParser(
            [parsers["gamemode"], InterpolationParser("gamemode")]
        ),
        "message": AlternativeParser(
            [FinalExpressionParser(InterpolationParser("message")), parsers["message"]]
        ),
        "block_pos": AlternativeParser(
            [parsers["block_pos"], InterpolationParser("vec3")]
        ),
        "column_pos": AlternativeParser(
            [parsers["column_pos"], InterpolationParser("vec2")]
        ),
        "rotation": AlternativeParser(
            [parsers["rotation"], InterpolationParser("vec2")]
        ),
        "vec2": AlternativeParser([parsers["vec2"], InterpolationParser("vec2")]),
        "vec3": AlternativeParser([parsers["vec3"], InterpolationParser("vec3")]),
        "entity": CommentDisambiguation(
            AlternativeParser([InterpolationParser("entity"), parsers["entity"]])
        ),
        "score_holder": CommentDisambiguation(
            AlternativeParser([InterpolationParser("entity"), parsers["score_holder"]])
        ),
        "game_profile": CommentDisambiguation(
            AlternativeParser([InterpolationParser("entity"), parsers["game_profile"]])
        ),
    }


def get_stream_lexical_scope(stream: TokenStream) -> LexicalScope:
    """Return the lexical scope currently associated with the token stream."""
    return stream.data.setdefault("lexical_scope", LexicalScope())


def get_stream_macro_scope(stream: TokenStream) -> Dict[str, AstMacro]:
    """Return the macro identifiers currently available."""
    return stream.data.setdefault("macro_scope", {})


def get_stream_pending_macros(stream: TokenStream) -> List[AstMacro]:
    """Return pending macro declarations."""
    return stream.data.setdefault("pending_macros", [])


@dataclass
class ToplevelHandler:
    """Handle toplevel root node."""

    parser: Parser

    modules: ModuleManager
    macro_handler: "MacroHandler"

    def __call__(self, stream: TokenStream) -> Any:
        current = self.modules.database.current
        resource_location = self.modules.database[current].resource_location

        prelude = self.modules.load_prelude()

        if module := self.modules.registry.get(current):
            return module.ast

        global_scope = GlobalScope(
            identifiers=set(self.modules.globals)
            | self.modules.builtins
            | {"__name__"},
        )

        lexical_scope = global_scope.push(LexicalScope)
        self.modules.parse_scopes[current] = lexical_scope

        pending_macros = get_stream_pending_macros(stream)

        for from_import in prelude:
            for item in from_import.arguments[1:]:
                if isinstance(item, AstImportedMacro):
                    pending_macros.append(item.declaration)
                elif isinstance(item, AstImportedItem):
                    lexical_scope.bind_variable(item.name, item)

        with (
            self.modules.parse_push(current),
            stream.provide(
                resource_location=resource_location,
                lexical_scope=lexical_scope,
            ),
        ):
            node = self.parser(stream)

            if prelude and isinstance(node, AstRoot):
                node = replace(node, commands=AstChildren([*prelude, *node.commands]))

        self.macro_handler.cache_local_spec(stream)

        if type(node) is AstRoot and not isinstance(current, Function):
            node = set_location(AstNonFunctionRoot(commands=node.commands), node)

        return node


def create_bolt_root_parser(parser: Parser, macro_handler: "MacroHandler"):
    """Compose root parsers."""
    parser = FlushPendingBindingsParser(parser, before=True)
    parser = DocstringHandler(parser)
    parser = VanillaReturnHandler(parser)
    parser = IfElseLoweringParser(parser)
    parser = BreakContinueConstraint(
        parser,
        allowed_scopes={
            ("while", "condition", "body"),
            ("for", "target", "in", "iterable", "body"),
        },
    )
    parser = LexicalScopeConstraint(
        parser,
        type=FunctionScope,
        flags={"memo": False},
        command_identifiers={
            "return",
            "return:pythonresult",
            "yield",
            "yield:value",
            "yield:from:value",
            "global:subcommand",
            "nonlocal:subcommand",
        },
    )
    parser = DeferredRootBacktracker(parser, macro_handler=macro_handler)
    parser = EscapeAnalysisResolver(parser)
    parser = DecoratorResolver(parser)
    parser = ProcMacroExpansion(parser)
    parser = RootScopeHandler(parser)
    return parser


def create_bolt_command_parser(
    parser: Parser, modules: ModuleManager, bolt_prototypes: Set[str]
):
    """Compose command parsers."""
    parser = MemoHandler(parser)
    parser = BindingStorageHandler(parser)
    parser = ImportStatementHandler(parser, modules)
    parser = SubcommandConstraint(parser, command_identifiers=bolt_prototypes)
    return parser


@dataclass
class ImprovedErrorHandler:
    """Parser that tries to recover more specific error alternatives."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        try:
            return self.parser(stream)
        except UndefinedIdentifier:
            raise
        except (UnexpectedToken, UnexpectedEOF) as exc:
            for exc_type, alts in exc.alternatives.items():
                if not issubclass(exc_type, (UnexpectedToken, UnexpectedEOF)):
                    for alt in alts:
                        if alt.end_location.pos + 1 >= exc.location.pos:
                            if isinstance(alt, UndefinedIdentifier):
                                alt.notes.append(str(exc))
                            raise alt from None
            raise


def parse_statement(stream: TokenStream) -> AstStatement:
    """Parse statement."""
    node = delegate("bolt:statement", stream)
    return set_location(AstStatement(arguments=AstChildren([node])), node)


@dataclass
class BranchScopeManager:
    """Parser that manages accessible identifiers for conditional branches."""

    parser: Parser
    init: bool = False
    fork: bool = False

    def __call__(self, stream: TokenStream) -> Any:
        lexical_scope = get_stream_lexical_scope(stream)

        if self.init or not lexical_scope.next_branch:
            lexical_scope.next_branch = lexical_scope.fork()

        branch_scope = lexical_scope.next_branch
        if self.fork:
            branch_scope = branch_scope.fork()

        with stream.provide(lexical_scope=branch_scope):
            node = self.parser(stream)

        if self.fork:
            lexical_scope.reconcile(branch_scope)

        return node


@dataclass
class FinalExpressionParser:
    """Parser that verifies that the expression isn't followed by anything."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        node = self.parser(stream)

        current_index = stream.index

        if consume_line_continuation(stream):
            exc = InvalidSyntax("Invalid indent following final expression.")
            raise set_location(exc, stream.get())

        next_token = stream.get()
        if next_token and not next_token.match("newline", "eof"):
            exc = InvalidSyntax("Trailing input following final expression.")
            raise set_location(exc, next_token)

        stream.index = current_index

        return node


@dataclass
class InterpolationParser:
    """Parser for interpolation."""

    converter: str
    prefix: Optional[str] = None
    unpack: Optional[str] = None

    def __call__(self, stream: TokenStream) -> Any:
        if stream.data.get("disable_interpolation"):
            token = stream.expect()
            raise set_location(InvalidSyntax("Interpolation disabled."), token)

        with stream.syntax(prefix=self.prefix, unpack=self.unpack):
            prefix = stream.get("prefix")
            unpack = self.unpack is not None and stream.expect("unpack")

        node = delegate("bolt:interpolation", stream)
        node = AstInterpolation(
            prefix=prefix.value if prefix else None,
            unpack=unpack.value if unpack else None,
            converter=self.converter,
            value=node,
        )
        return set_location(node, prefix or node.value, node.value)


@dataclass
class DisableInterpolationParser:
    """Parser that disables interpolation."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        with stream.provide(disable_interpolation=True):
            return self.parser(stream)


@dataclass
class ForkLexicalScopeParser:
    """Parser forking lexical scope."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        lexical_scope = get_stream_lexical_scope(stream)
        temporary_scope = lexical_scope.fork()
        with stream.provide(lexical_scope=temporary_scope):
            node = self.parser(stream)
        lexical_scope.reconcile(temporary_scope)
        return node


@dataclass
class AssignmentStatementParser:
    """Parser for assignment statements."""

    parser: Parser
    binding_only: bool = False

    def __call__(self, stream: TokenStream) -> Any:
        node = self.parser(stream)

        assignment_pattern = r"=(?!=)|\+=|-=|\*=|//=|/=|%=|&=|\|=|\^=|<<=|>>=|\*\*="

        if isinstance(node, AstTarget):
            type_annotation = None
            identifier = None
            if isinstance(node, AstTargetIdentifier):
                with stream.syntax(colon=r":"):
                    if stream.get("colon"):
                        type_annotation = delegate("bolt:type_annotation", stream)
                        identifier = node
                        assignment_pattern = r"=(?!=)"

            if self.binding_only:
                assignment_pattern = r"=(?!=)"
            with stream.syntax(assignment=assignment_pattern):
                if identifier and type_annotation:
                    op = stream.get("assignment")
                    if not op:
                        lexical_scope = get_stream_lexical_scope(stream)
                        for i, (_, node) in reversed(
                            list(enumerate(lexical_scope.pending_bindings))
                        ):
                            if node is identifier:
                                del lexical_scope.pending_bindings[i]
                                break
                        node = AstTypeDeclaration(
                            identifier=identifier,
                            type_annotation=type_annotation,
                        )
                        return set_location(node, identifier, type_annotation)
                else:
                    op = stream.expect("assignment")

            expression = delegate("bolt:expression", stream)

            node = AstAssignment(
                operator=op.value,
                target=node,
                value=expression,
                type_annotation=type_annotation,
            )
            node = set_location(node, node.target, node.value)

        elif isinstance(node, (AstAttribute, AstLookup)):
            with stream.syntax(assignment=assignment_pattern):
                op = stream.get("assignment")

            if op:
                expression = delegate("bolt:expression", stream)
                target = (
                    AstTargetAttribute(name=node.name, value=node.value)
                    if isinstance(node, AstAttribute)
                    else AstTargetItem(value=node.value, arguments=node.arguments)
                )
                node = AstAssignment(
                    operator=op.value,
                    target=set_location(target, node),
                    value=expression,
                )
                node = set_location(node, node.target, node.value)

        return node


@dataclass
class EscapeAnalysisParser:
    """Initiate escape analysis on a nested block."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        lexical_scope = get_stream_lexical_scope(stream)
        previous_bindings = {
            identifier: len(variable.bindings)
            for identifier, variable in lexical_scope.variables.items()
        }

        if isinstance(node := self.parser(stream), AstRoot):
            refcount_snapshots: List[Tuple[str, Binding, int]] = []

            for identifier, variable in lexical_scope.variables.items():
                nb_bindings = previous_bindings.get(identifier)
                last_binding = len(variable.bindings) - 1
                if nb_bindings is None or last_binding >= nb_bindings:
                    binding = variable.bindings[last_binding]
                    refcount = len(binding.references)
                    if variable.storage != "local":
                        exc = InvalidSyntax(
                            f'Binding to {variable.storage} variable "{identifier}" is not allowed in this context.'
                        )
                        raise set_location(exc, binding.origin)
                    refcount_snapshots.append((identifier, binding, refcount))

            if refcount_snapshots:
                escape_analysis = AstEscapeAnalysisRoot(
                    commands=node.commands,
                    refcount_snapshots=refcount_snapshots,
                )
                node = set_location(escape_analysis, node)
                stream.data["do_escape_analysis"] = True

        return node


@dataclass
class EscapeAnalysisResolver:
    """Resolve escape analysis."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        if not stream.data.get("root_scope"):
            return self.parser(stream)

        stream.data["do_escape_analysis"] = False

        if (
            isinstance(node := self.parser(stream), AstRoot)
            and stream.data["do_escape_analysis"]
        ):
            node = self.resolve(node)

        return node

    def resolve(self, node: AstRoot) -> AstRoot:
        should_replace = False
        commands: List[AstCommand] = []

        for command in node.commands:
            stack: List[AstCommand] = [command]

            while command.arguments and isinstance(
                command := command.arguments[-1], AstCommand
            ):
                stack.append(command)

            command = stack.pop()

            if (
                command.arguments
                and isinstance(root := command.arguments[-1], AstRoot)
                and (resolved_root := self.resolve(root)) is not root
            ):
                should_replace = True

                command = replace(
                    command,
                    arguments=AstChildren([*command.arguments[:-1], resolved_root]),
                )
                while stack:
                    parent = stack.pop()
                    command = replace(
                        parent,
                        arguments=AstChildren([*parent.arguments[:-1], command]),
                    )

            elif stack:
                command = stack[0]

            commands.append(command)

        if isinstance(node, AstEscapeAnalysisRoot):
            identifiers = tuple(
                name
                for name, binding, refcount in node.refcount_snapshots
                if len(binding.references) > refcount
            )
            node = set_location(
                (
                    AstEscapeRoot(
                        commands=AstChildren(commands), identifiers=identifiers
                    )
                    if identifiers
                    else AstRoot(commands=AstChildren(commands))
                ),
                node,
            )
        elif should_replace:
            node = replace(node, commands=AstChildren(commands))

        return node


@dataclass
class MemoRootParser:
    """Parse memo root."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        with stream.provide(loop=False, memo=True):
            return self.parser(stream)


@dataclass
class MemoHandler:
    """Handle memo."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        node = self.parser(stream)

        if not isinstance(node, AstCommand):
            return node
        if node.identifier == "memo:subcommand":
            node = self.create_memo(node, stream.data.get("uuid_factory", uuid4)())

        return node

    def create_memo(self, command: AstCommand, persistent_id: UUID) -> AstMemo:
        identifier_parts: List[str] = ["memo"]
        arguments: List[AstNode] = []

        node = command
        while isinstance(subcommand := node.arguments[-1], AstCommand):
            if isinstance(assignment := node.arguments[0], AstAssignment):
                if not isinstance(assignment.target, AstTargetIdentifier):
                    exc = InvalidSyntax("Assignment target must be an identifier.")
                    raise set_location(exc, assignment.target)
                identifier_parts.append(assignment.target.value)
                arguments.append(assignment)
            elif isinstance(expression := node.arguments[0], AstExpression):
                identifier_parts.append(str(len(arguments)))
                arguments.append(expression)
            node = subcommand

        arguments.append(node.arguments[0])

        memo = AstMemo(
            identifier=":".join(identifier_parts),
            arguments=AstChildren(arguments),
            persistent_id=persistent_id,
        )

        return set_location(memo, command)


def parse_decorator(stream: TokenStream) -> Any:
    """Parse decorator."""
    with stream.syntax(decorator="@"):
        token = stream.expect("decorator")
        expression = delegate("bolt:expression", stream)
        return set_location(AstDecorator(expression=expression), token, expression)


@dataclass
class DecoratorResolver:
    """Parser for resolving decorators."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> AstRoot:
        node: AstRoot = self.parser(stream)

        stack: List[AstDecorator] = []

        changed = False
        result: List[AstCommand] = []

        for command in node.commands:
            if isinstance(command, AstStatement) and isinstance(
                decorator := command.arguments[0], AstDecorator
            ):
                changed = True
                stack.append(decorator)
            elif stack and isinstance(
                target := command.arguments[0], (AstFunctionSignature, AstClassName)
            ):
                arg = replace(target, decorators=AstChildren(stack))
                result.append(
                    replace(
                        command, arguments=AstChildren([arg, *command.arguments[1:]])
                    )
                )
                stack.clear()
            elif stack:
                message = "Decorators can only be applied to functions or classes."
                raise set_location(InvalidSyntax(message), stack[-1])
            else:
                result.append(command)

        if stack:
            exc = InvalidSyntax("Can't apply decorator to nothing.")
            raise set_location(exc, stack[-1])

        if changed:
            node = replace(node, commands=AstChildren(result))

        return node


@dataclass
class AssignmentTargetParser:
    """Parser for assignment targets."""

    require_local_binding: bool = False
    require_single: bool = False

    def __call__(self, stream: TokenStream) -> AstTarget:
        lexical_scope = get_stream_lexical_scope(stream)

        nodes: List[AstTarget] = []

        with stream.syntax(identifier=IDENTIFIER_PATTERN, comma=r","):
            while True:
                token = stream.expect("identifier")
                rebind = lexical_scope.has_binding(token.value)

                target = AstTargetIdentifier(value=token.value, rebind=rebind)
                target = set_location(target, token)

                if self.require_local_binding:
                    lexical_scope.reference_binding(token.value, target)

                lexical_scope.create_pending_binding(token.value, target)

                nodes.append(target)

                if self.require_single or not stream.get("comma"):
                    break

        if len(nodes) == 1:
            return nodes[0]

        node = AstTargetUnpack(targets=AstChildren(nodes))
        return set_location(node, nodes[0], nodes[-1])


@dataclass
class VanillaReturnHandler:
    """Turn vanilla return into bolt return statement when not prefixed by execute."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        node: AstRoot = self.parser(stream)

        lexical_scope = get_stream_lexical_scope(stream)
        if not isinstance(lexical_scope, FunctionScope):
            return node

        changed = False
        result: List[AstCommand] = []

        for command in node.commands:
            if command.identifier == "return:value" and command.arguments:
                changed = True

                arg = command.arguments[0]
                if isinstance(arg, AstNumber):
                    arg = set_location(AstValue(value=arg.value), arg)
                elif isinstance(arg, AstInterpolation):
                    arg = arg.value

                command = set_location(
                    AstCommand(
                        identifier="return:pythonresult", arguments=AstChildren([arg])
                    ),
                    command,
                )

            result.append(command)

        if changed:
            node = replace(node, commands=AstChildren(result))

        return node


@dataclass
class IfElseLoweringParser:
    """Parser that turns elif statements into if statements in an else clause."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> AstRoot:
        node: AstRoot = self.parser(stream)

        changed = False
        result: List[AstCommand] = []

        commands = iter(node.commands)
        previous = ""

        for command in commands:
            if command.identifier in ["elif:condition:body", "else:body"]:
                if previous not in ["if:condition:body", "elif:condition:body"]:
                    exc = InvalidSyntax(
                        "Conditional branch must be part of an if statement."
                    )
                    raise set_location(exc, command, command.arguments[-1].location)

            if command.identifier == "elif:condition:body":
                changed = True
                elif_chain = [command]

                for command in commands:
                    if command.identifier not in ["elif:condition:body", "else:body"]:
                        break
                    elif_chain.append(command)
                else:
                    command = None

                if elif_chain[-1].identifier == "else:body":
                    last = [elif_chain.pop()]
                else:
                    last = []

                for stmt in reversed(elif_chain):
                    stmt = replace(stmt, identifier="if:condition:body")
                    stmt = AstRoot(commands=AstChildren([stmt, *last]))
                    stmt = set_location(stmt, stmt.commands[0], stmt.commands[-1])
                    stmt = AstCommand(
                        identifier="else:body",
                        arguments=AstChildren([stmt]),
                    )
                    last = [set_location(stmt, stmt.arguments[0])]

                result.append(last[0])

            if not command:
                break

            previous = command.identifier
            result.append(command)

        if changed:
            node = replace(node, commands=AstChildren(result))

        return node


@dataclass
class BreakContinueConstraint:
    """Constraint that makes sure that break and continue statements only occur in loops."""

    parser: Parser
    allowed_scopes: Set[Tuple[str, ...]]

    def __call__(self, stream: TokenStream) -> AstRoot:
        scope = get_stream_scope(stream)
        loop = stream.data.get("loop") or scope in self.allowed_scopes

        with stream.provide(loop=loop):
            node: AstRoot = self.parser(stream)

        if not loop:
            for command in node.commands:
                if command.identifier in ["break", "continue"]:
                    exc = InvalidSyntax(
                        f'Can only use "{command.identifier}" in loops.'
                    )
                    raise set_location(exc, command)

        return node


def parse_deferred_root(stream: TokenStream) -> AstDeferredRoot:
    """Parse deferred root."""
    stream_copy = stream.copy()

    with stream.syntax(colon=r":"):
        token = stream.expect("colon")

    with stream.syntax(statement=r"[^\s#].*"):
        while consume_line_continuation(stream):
            stream.expect("statement")

    lexical_scope = get_stream_lexical_scope(stream)

    deferred_scope = lexical_scope.deferred(LexicalScope)
    stream_copy.data["lexical_scope"] = deferred_scope
    del stream_copy.data["root_scope"]
    lexical_scope.deferred_complete()

    node = AstDeferredRoot(stream=stream_copy)
    return set_location(node, token, stream.current)


def parse_function_signature(stream: TokenStream) -> AstFunctionSignature:
    """Parse function signature."""
    lexical_scope = get_stream_lexical_scope(stream)
    type_annotation_scope = lexical_scope.fork()

    arguments: List[AstFunctionSignatureElement] = []

    encountered_positional = False
    encountered_default = False
    encountered_variadic = False
    encountered_variadic_keyword = False
    expect_named_argument = False

    with stream.syntax(
        comma=r",",
        equal=r"=(?!=)",
        brace=r"\(|\)",
        separator=r"/",
        kwargs=r"\*\*",
        args=r"\*",
        colon=r":",
        arrow=r"->",
        identifier=IDENTIFIER_PATTERN,
    ):
        identifier = stream.expect("identifier")
        stream.expect(("brace", "("))

        node = set_location(AstFunctionSignature(name=identifier.value), identifier)
        lexical_scope.bind_variable(identifier.value, node)

        deferred_scope = lexical_scope.deferred(FunctionScope)

        with stream.ignore("newline"):
            for token in stream.peek_until(("brace", ")")):
                if encountered_variadic_keyword:
                    exc = InvalidSyntax(
                        "Variadic keyword arguments should not be followed by anything."
                    )
                    raise set_location(exc, token)

                name, separator, kwargs, args = stream.expect(
                    "identifier",
                    "separator",
                    "kwargs",
                    "args",
                )

                if name:
                    expect_named_argument = False

                    type_annotation = None
                    if stream.get("colon"):
                        with stream.provide(lexical_scope=type_annotation_scope):
                            type_annotation = delegate("bolt:type_annotation", stream)

                    default = None
                    if stream.get("equal"):
                        encountered_default = True
                        with stream.provide(lexical_scope=deferred_scope):
                            default = delegate("bolt:expression", stream)
                    elif encountered_default:
                        exc = InvalidSyntax(
                            "Argument without default can not appear after arguments with default values."
                        )
                        raise set_location(exc, token)

                    argument = AstFunctionSignatureArgument(
                        name=name.value,
                        default=default,
                        type_annotation=type_annotation,
                    )
                    argument = set_location(argument, name, stream.current)
                    deferred_scope.bind_variable(argument.name, argument)
                    arguments.append(argument)

                elif separator:
                    if encountered_positional:
                        exc = InvalidSyntax(
                            "Positional marker already present in function signature."
                        )
                        raise set_location(exc, separator)
                    if encountered_variadic:
                        exc = InvalidSyntax(
                            "Positional marker can not appear after variadic arguments."
                        )
                        raise set_location(exc, separator)
                    if not arguments:
                        exc = InvalidSyntax(
                            "Positional marker can not appear directly at the beginning of the argument list."
                        )
                        raise set_location(exc, separator)
                    encountered_positional = True
                    argument = AstFunctionSignaturePositionalMarker()
                    arguments.append(set_location(argument, separator))

                elif kwargs:
                    encountered_variadic_keyword = True
                    name = stream.expect("identifier")

                    type_annotation = None
                    if stream.get("colon"):
                        with stream.provide(lexical_scope=type_annotation_scope):
                            type_annotation = delegate("bolt:type_annotation", stream)

                    argument = AstFunctionSignatureVariadicKeywordArgument(
                        name=name.value,
                        type_annotation=type_annotation,
                    )
                    argument = set_location(argument, kwargs, name)
                    deferred_scope.bind_variable(argument.name, argument)
                    arguments.append(argument)

                elif args:
                    if encountered_variadic:
                        exc = InvalidSyntax("Variadic arguments already specified.")
                        raise set_location(exc, args)
                    encountered_variadic = True
                    if name := stream.get("identifier"):
                        type_annotation = None
                        if stream.get("colon"):
                            with stream.provide(lexical_scope=type_annotation_scope):
                                type_annotation = delegate(
                                    "bolt:type_annotation", stream
                                )

                        argument = AstFunctionSignatureVariadicArgument(
                            name=name.value,
                            type_annotation=type_annotation,
                        )
                        argument = set_location(argument, args, name)
                        deferred_scope.bind_variable(argument.name, argument)
                        arguments.append(argument)
                    else:
                        expect_named_argument = True
                        argument = AstFunctionSignatureVariadicMarker()
                        arguments.append(set_location(argument, args))

                if not stream.get("comma"):
                    stream.expect(("brace", ")"))
                    break

            if expect_named_argument:
                for argument in arguments:
                    if isinstance(argument, AstFunctionSignatureVariadicMarker):
                        exc = InvalidSyntax(
                            "Expected at least one named argument after bare variadic marker."
                        )
                        raise set_location(exc, argument)

            return_type_annotation = None
            if stream.get("arrow"):
                with stream.provide(lexical_scope=type_annotation_scope):
                    return_type_annotation = delegate("bolt:type_annotation", stream)

    node = replace(
        node,
        arguments=AstChildren(arguments),
        return_type_annotation=return_type_annotation,
    )
    return set_location(node, node, stream.current)


def parse_proc_macro_signature(stream: TokenStream):
    """Parse proc macro signature."""

    with stream.syntax(brace=r"\(|\)", identifier=IDENTIFIER_PATTERN):
        begin = stream.expect(("brace", "("))
        stream.expect(("identifier", "stream"))
        end = stream.expect(("brace", ")"))

    node = set_location(AstProcMacroMarker(), begin, end)

    lexical_scope = get_stream_lexical_scope(stream)
    deferred_scope = lexical_scope.deferred(ProcMacroScope)
    deferred_scope.bind_variable("stream", node)

    return node


@dataclass
class MacroMatchParser:
    """Parser for macro matching."""

    literal_parser: Parser
    argument_parser: Parser
    resource_location_parser: Parser
    json_properties_parser: Parser

    def __call__(self, stream: TokenStream) -> AstMacroMatch:
        spec = get_stream_spec(stream)

        with stream.syntax(equal=r"(?<!\s)=(?!\s)"), stream.checkpoint() as commit:
            argument = self.argument_parser(stream)
            stream.expect("equal")
            commit()

            node = AstMacroMatchArgument(
                match_identifier=argument,
                match_argument_parser=self.resource_location_parser(stream),
                match_argument_properties=self.json_properties_parser(stream),
            )
            node = set_location(
                node,
                argument,
                node.match_argument_properties or node.match_argument_parser,
            )

            if not node.is_subcommand():
                name = node.match_argument_parser.get_canonical_value()
                if f"command:argument:{name}" not in spec.parsers:
                    exc = InvalidSyntax(f'Unrecognized argument parser "{name}".')
                    raise set_location(exc, node.match_argument_parser)

            lexical_scope = get_stream_lexical_scope(stream)
            deferred_scope = lexical_scope.deferred(MacroScope)
            deferred_scope.bind_variable(node.match_identifier.value, node)

        if commit.rollback:
            literal = self.literal_parser(stream)
            node = set_location(AstMacroMatchLiteral(match=literal), literal)

        return node


@dataclass
class MacroHandler:
    """Handle macros."""

    parser: Parser

    proc_macro_overloads: Dict[str, Dict[str, int]] = extra_field(default_factory=dict)
    spec_cache: Dict[FrozenSet[AstMacro], CommandSpec] = extra_field(
        default_factory=dict
    )

    def __call__(self, stream: TokenStream) -> Any:
        should_flush = not (
            scope[0] in ["macro", "from"]
            if (scope := get_stream_scope(stream))
            else (
                (token := stream.peek())
                and token.match(("literal", "macro"), ("literal", "from"))
            )
        )
        if should_flush:
            self.flush_pending_macros(stream)

        node = self.parser(stream)

        if not isinstance(node, AstCommand):
            return node

        if node.identifier == "macro:name:subcommand":
            self.check_root_scope(stream, node)
            node = self.create_macro(node, stream.data.get("resource_location"))
            if not isinstance(node, AstProcMacro):
                declaration = replace(node, arguments=AstChildren(node.arguments[:-1]))
                get_stream_pending_macros(stream).append(declaration)

        elif macro := get_stream_macro_scope(stream).get(node.identifier):
            if not isinstance(macro, AstProcMacro):
                node = set_location(
                    AstMacroCall(identifier=node.identifier, arguments=node.arguments),
                    node,
                )

        return node

    def check_root_scope(self, stream: TokenStream, node: AstCommand):
        if not stream.data.get("root_scope"):
            message = "Macro definition can only appear directly at scope level."
            raise set_location(InvalidSyntax(message), node, node.arguments[0])

    def create_macro(
        self,
        command: AstCommand,
        resource_location: Optional[str] = None,
    ) -> AstMacro:
        identifier_parts: List[str] = []
        arguments: List[AstNode] = []

        node = command
        while isinstance(subcommand := node.arguments[-1], AstCommand):
            if isinstance(literal := node.arguments[0], AstMacroLiteral):
                if literal.value in ["macro", "from"]:
                    exc = InvalidSyntax(f'Forbidden literal "{literal.value}".')
                    raise set_location(exc, literal)
                identifier_parts.append(literal.value)
            elif isinstance(literal := node.arguments[0], AstMacroMatchLiteral):
                identifier_parts.append(literal.match.value)
            elif isinstance(argument := node.arguments[0], AstMacroMatchArgument):
                identifier_parts.append(
                    "subcommand"
                    if argument.is_subcommand()
                    else argument.match_identifier.value
                )
            arguments.append(node.arguments[0])
            node = subcommand

        if isinstance(marker := node.arguments[0], AstProcMacroMarker):
            identifier = ":".join(identifier_parts)

            if not resource_location:
                message = f'Can\'t define proc macro "{identifier_parts[0]}" without resource location information.'
                raise set_location(InvalidSyntax(message), arguments[0], marker)

            for argument in arguments:
                if isinstance(argument, AstMacroMatchArgument):
                    message = f'Proc macro "{identifier_parts[0]}" contains non-literal match.'
                    raise set_location(InvalidSyntax(message), argument)

            macro_overload = self.proc_macro_overloads.setdefault(resource_location, {})
            overload_id = macro_overload.setdefault(identifier, 0)
            macro_overload[identifier] += 1

            argument = f"proc_macro_overload{overload_id}"
            identifier = f"{identifier}:{argument}"

            match_node = AstMacroMatchArgument(
                match_identifier=AstMacroArgument(value=argument),
                match_argument_parser=AstResourceLocation.from_value("bolt:proc_macro"),
                match_argument_properties=AstJson.from_value(
                    {
                        "resource_location": resource_location,
                        "identifier": identifier,
                    }
                ),
            )
            arguments.append(match_node)
            arguments.append(node.arguments[-1])

            macro = AstProcMacro(
                identifier=identifier,
                arguments=AstChildren(arguments),
            )
            return set_location(macro, command)

        for left, right in zip(arguments, arguments[1:]):
            if isinstance(left, AstMacroMatchArgument) and left.is_subcommand():
                exc = InvalidSyntax("Unexpected macro argument after subcommand.")
                raise set_location(exc, right)

        arguments.append(node.arguments[0])

        macro = AstMacro(
            identifier=":".join(identifier_parts),
            arguments=AstChildren(arguments),
        )

        return set_location(macro, command)

    def flush_pending_macros(self, stream: TokenStream):
        if pending_macros := get_stream_pending_macros(stream):
            self.inject_syntax(stream, *pending_macros)
            pending_macros.clear()

    def cache_local_spec(self, stream: TokenStream):
        if stream.data.get("local_spec"):
            scope_key = frozenset(get_stream_macro_scope(stream).values())
            self.spec_cache[scope_key] = get_stream_spec(stream)
            stream.data["local_spec"] = False

    def inject_syntax(self, stream: TokenStream, *macros: AstMacro):
        macro_scope = get_stream_macro_scope(stream)
        scope_key = frozenset(macros) | frozenset(macro_scope.values())

        if len(macro_scope) == len(scope_key):
            return

        if spec := self.spec_cache.get(scope_key):
            stream.data["local_spec"] = False
            stream.data["spec"] = spec
            stream.data["macro_scope"] = {
                macro.identifier: macro for macro in scope_key
            }
            return

        spec = get_stream_spec(stream)

        if not stream.data.get("local_spec"):
            stream.data["local_spec"] = True
            spec = replace(spec, tree=spec.tree.copy(deep=True), prototypes={})
            stream.data["spec"] = spec
            macro_scope = macro_scope.copy()
            stream.data["macro_scope"] = macro_scope

        for macro in macros:
            spec.tree.extend(CommandTree.parse_obj(macro.get_command_tree()))
            macro_scope[macro.identifier] = macro

        spec.update()


@dataclass
class ProcMacroParser:
    """Parser for invoking proc macros."""

    modules: ModuleManager

    @internal
    def __call__(self, stream: TokenStream) -> AstProcMacroResult:
        properties = get_stream_properties(stream)
        resource_location: str = properties["resource_location"]
        identifier: str = properties["identifier"]

        module = self.modules[resource_location]

        if not module.executed:
            with self.modules.error_handler(
                "Top-level statement raised an exception.",
                resource_location,
            ):
                self.modules.eval(module)

        runtime: CommandEmitter = module.namespace["_bolt_runtime"]
        macro = module.namespace["_bolt_proc_macros"][identifier]

        with self.modules.error_handler(
            f'Proc macro "{identifier}" raised an exception.', resource_location
        ):
            result = runtime.capture_output(macro, stream)

        return AstProcMacroResult(commands=AstChildren(result))


@dataclass
class ProcMacroExpansion:
    """Expand proc macro results."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> AstRoot:
        should_replace = False
        commands: List[AstCommand] = []

        node: AstRoot = self.parser(stream)

        for command in node.commands:
            stack: List[AstCommand] = [command]

            while command.arguments and isinstance(
                command := command.arguments[-1], AstCommand
            ):
                stack.append(command)

            command = stack.pop()

            if command.arguments and isinstance(
                result := command.arguments[-1], AstProcMacroResult
            ):
                should_replace = True

                for expansion in result.commands:
                    for prefix in reversed(stack):
                        args = AstChildren([*prefix.arguments[:-1], expansion])
                        expansion = replace(prefix, arguments=args)
                    commands.append(expansion)

                continue

            elif stack:
                command = stack[0]

            commands.append(command)

        if should_replace:
            return replace(node, commands=AstChildren(commands))

        return node


def parse_class_name(stream: TokenStream) -> AstClassName:
    """Parse class name."""
    with stream.syntax(identifier=IDENTIFIER_PATTERN):
        token = stream.expect("identifier")
        node = set_location(AstClassName(value=token.value), token)

    lexical_scope = get_stream_lexical_scope(stream)
    lexical_scope.create_pending_binding(node.value, node)

    return node


@dataclass
class ClassBasesParser:
    """Parser for class bases."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> AstClassBases:
        inherit: List[AstExpression] = []
        kwargs: List[AstKeyword] = []

        with stream.syntax(brace=r"\(|\)", comma=","):
            token = stream.expect(("brace", "("))

            with stream.ignore("newline"):
                for _ in stream.peek_until(("brace", ")")):
                    arg = self.parser(stream)

                    if isinstance(arg, AstKeyword):
                        kwargs.append(arg)
                    else:
                        if kwargs:
                            exc = InvalidSyntax(
                                f'Base class not allowed after keyword argument "{kwargs[-1].name}".'
                            )
                            raise set_location(exc, arg)
                        inherit.append(arg)

                    if not stream.get("comma"):
                        stream.expect(("brace", ")"))
                        break

        node = AstClassBases(inherit=AstChildren(inherit), kwargs=AstChildren(kwargs))
        return set_location(node, token, stream.current)


def parse_class_root(stream: TokenStream) -> AstClassRoot:
    """Parse class root."""
    lexical_scope = get_stream_lexical_scope(stream)

    with stream.provide(lexical_scope=lexical_scope.push(ClassScope)):
        if isinstance(node := delegate("nested_root", stream), AstRoot):
            node = set_location(AstClassRoot(commands=node.commands), node)

    return node


def parse_del_target(stream: TokenStream) -> AstTarget:
    """Parse del target."""
    node = delegate("bolt:expression", stream)

    if isinstance(node, AstIdentifier):
        return set_location(AstTargetIdentifier(value=node.value), node)
    elif isinstance(node, AstAttribute):
        return set_location(AstTargetAttribute(name=node.name, value=node.value), node)
    elif isinstance(node, AstLookup):
        return set_location(
            AstTargetItem(value=node.value, arguments=node.arguments), node
        )

    exc = InvalidSyntax("Can only delete variables, attributes, or subscripted items.")
    raise set_location(exc, node)


def parse_identifier(stream: TokenStream) -> AstIdentifier:
    """Parse identifier."""
    lexical_scope = get_stream_lexical_scope(stream)

    with stream.syntax(
        true=TRUE_PATTERN,
        false=FALSE_PATTERN,
        null=NULL_PATTERN,
        identifier=IDENTIFIER_PATTERN,
    ):
        token = stream.expect("identifier")
        node = set_location(AstIdentifier(value=token.value), token)

    lexical_scope.reference_variable(node.value, node)

    return node


@dataclass
class TrailingCommaParser:
    """Parser for discarding trailing comma."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        node = self.parser(stream)
        with stream.syntax(comma=r","):
            stream.get("comma")
        return node


@dataclass
class ImportLocationConstraint:
    """Constraint for import location."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        node: AstResourceLocation = self.parser(stream)

        if node.is_tag or node.namespace is None and not IMPORT_REGEX.match(node.path):
            exc = InvalidSyntax(f'Invalid module location "{node.get_value()}".')
            raise set_location(exc, node)

        return node


@dataclass
class ImportStatementHandler:
    """Handle import statements."""

    parser: Parser
    modules: ModuleManager

    def __call__(self, stream: TokenStream) -> Any:
        if isinstance(node := self.parser(stream), AstCommand):
            if node.identifier in ["import:module", "import:module:as:alias"]:
                self.check_root_scope(stream, node)
                return self.handle_import(stream, node)
            elif node.identifier == "from:module:import:subcommand":
                self.check_root_scope(stream, node)
                return self.handle_from_import(stream, node)
        return node

    def check_root_scope(self, stream: TokenStream, node: AstCommand):
        if not stream.data.get("root_scope"):
            message = "Import statement can only appear directly at scope level."
            raise set_location(InvalidSyntax(message), node)

    def handle_import(self, stream: TokenStream, node: AstCommand) -> AstCommand:
        lexical_scope = get_stream_lexical_scope(stream)

        if node.identifier == "import:module":
            if isinstance(module := node.arguments[0], AstResourceLocation):
                if module.namespace:
                    message = f'Can\'t import "{module.get_value()}" without alias.'
                    raise set_location(InvalidSyntax(message), module)
                lexical_scope.bind_variable(module.path.partition(".")[0], module)

        elif node.identifier == "import:module:as:alias":
            if isinstance(alias := node.arguments[1], AstImportedItem):
                if not alias.identifier:
                    exc = InvalidSyntax(f'Invalid identifier "{alias.name}".')
                    raise set_location(exc, alias)
                lexical_scope.bind_variable(alias.name, alias)

        return node

    @internal
    def handle_from_import(self, stream: TokenStream, node: AstCommand) -> AstCommand:
        lexical_scope = get_stream_lexical_scope(stream)
        pending_macros = get_stream_pending_macros(stream)

        module = cast(AstResourceLocation, node.arguments[0])
        target_name = module.get_value()
        target = self.modules.database.index.get(target_name)

        try:
            compiled_module = (
                target
                and target not in self.modules.parse_stack
                and self.modules.get(target)
            )
        except UnusableCompilationUnit:
            compiled_module = None

        arguments: List[AstNode] = [module]

        subcommand = cast(AstCommand, node.arguments[1])
        while True:
            if isinstance(item := subcommand.arguments[0], AstImportedItem):
                if compiled_module and item.name in compiled_module.macros:
                    for name, macro in compiled_module.macros[item.name]:
                        imported_macro = AstImportedMacro(name=name, declaration=macro)
                        arguments.append(set_location(imported_macro, item))
                        pending_macros.append(macro)

                elif item.identifier:
                    arguments.append(item)
                    lexical_scope.bind_variable(item.name, item)

                else:
                    exc = InvalidSyntax(f'Invalid identifier "{item.name}".')
                    raise set_location(exc, item)

            if subcommand.identifier == "from:module:import:name:subcommand":
                subcommand = cast(AstCommand, subcommand.arguments[1])
            else:
                break

        return set_location(AstFromImport(arguments=AstChildren(arguments)), node)


def parse_python_import(stream: TokenStream) -> Any:
    """Parse python import."""
    with stream.syntax(module=rf"{MODULE_PATTERN}(?=\s)"):
        token = stream.expect("module")
        return set_location(AstResourceLocation(path=token.value), token)


def parse_import_name(stream: TokenStream) -> AstImportedItem:
    """Parse import name."""
    with stream.syntax(name=IDENTIFIER_PATTERN, literal=r"[^#:\s]+(?<!,)"):
        token = stream.get("name")

        if token:
            node = AstImportedItem(name=token.value)
        else:
            token = stream.expect("literal")
            node = AstImportedItem(name=token.value, identifier=False)

        return set_location(node, token)


def parse_name_list(stream: TokenStream) -> AstIdentifier:
    """Parse name list."""
    node = delegate("bolt:identifier", stream)
    with stream.syntax(comma=r","):
        stream.get("comma")
    return node


@dataclass
class BindingStorageHandler:
    """Handle binding storage."""

    parser: Parser

    storage_qualifiers: Dict[str, Literal["global", "nonlocal", "deleted"]] = field(
        default_factory=lambda: {
            "global:subcommand": "global",
            "nonlocal:subcommand": "nonlocal",
            "del:target": "deleted",
        }
    )

    def __call__(self, stream: TokenStream) -> Any:
        if isinstance(node := self.parser(stream), AstCommand):
            if storage := self.storage_qualifiers.get(node.identifier):
                lexical_scope = get_stream_lexical_scope(stream)

                if isinstance(target := node.arguments[0], AstTargetIdentifier):
                    lexical_scope.bind_storage(target.value, storage, target)

                elif isinstance(subcommand := node.arguments[0], AstCommand):
                    while True:
                        if isinstance(name := subcommand.arguments[0], AstIdentifier):
                            lexical_scope.bind_storage(name.value, storage, name)
                        if subcommand.identifier == f"{storage}:name:subcommand":
                            subcommand = cast(AstCommand, subcommand.arguments[1])
                        else:
                            break

        return node


@dataclass
class DocstringHandler:
    """Emit docstring as special node."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        node: AstRoot = self.parser(stream)

        changed = False
        result: List[AstCommand] = []

        for command in node.commands:
            if (
                isinstance(command, AstStatement)
                and command.arguments
                and isinstance(literal := command.arguments[0], AstValue)
                and isinstance(literal.value, str)
            ):
                changed = True
                command = set_location(
                    AstDocstring(arguments=AstChildren([literal])), command
                )

            result.append(command)

        if changed:
            node = replace(node, commands=AstChildren(result))

        return node


@dataclass
class WithContextParser:
    """Parser for with context."""

    expression_parser: Parser
    target_parser: Parser

    def __call__(self, stream: TokenStream) -> AstWithContext:
        clauses: List[AstWithClause] = []

        with stream.syntax(
            brace=r"\(|\)",
            comma=r",",
            target=r"\bas\b",
        ):
            open_brace = stream.get(("brace", "("))

            while True:
                value = self.expression_parser(stream)

                target = None
                if stream.get("target"):
                    target = self.target_parser(stream)

                node = AstWithClause(value=value, target=target)
                clauses.append(set_location(node, value, target))

                if stream.get("comma"):
                    if open_brace and stream.get(("brace", ")")):
                        break
                else:
                    if open_brace:
                        stream.expect(("brace", ")"))
                    break

        node = AstWithContext(clauses=AstChildren(clauses))
        return set_location(node, open_brace or clauses[0], clauses[-1])


@dataclass
class FlushPendingBindingsParser:
    """Parser that flushes pending bindings."""

    parser: Parser
    before: bool = False
    after: bool = False

    def __call__(self, stream: TokenStream) -> Any:
        if self.before:
            lexical_scope = get_stream_lexical_scope(stream)
            lexical_scope.flush_pending_bindings()
        try:
            node = self.parser(stream)
        except Exception:
            lexical_scope = get_stream_lexical_scope(stream)
            lexical_scope.pending_bindings.clear()
            raise
        if self.after:
            lexical_scope = get_stream_lexical_scope(stream)
            lexical_scope.flush_pending_bindings()
        return node


@dataclass
class SubcommandConstraint:
    """Prevent using the specified prototypes as general subcommands."""

    parser: Parser
    command_identifiers: Set[str]

    def __call__(self, stream: TokenStream) -> Any:
        if (
            isinstance(node := self.parser(stream), AstCommand)
            and node.arguments
            and isinstance(child := node.arguments[-1], AstCommand)
            and child.identifier in self.command_identifiers
            and node.identifier not in self.command_identifiers
        ):
            name, _, _ = child.identifier.partition(":")
            exc = InvalidSyntax(f"Can't use {name} as a subcommand.")
            raise set_location(exc, child)
        return node


@dataclass
class DeferredRootBacktracker:
    """Parser for backtracking over deferred root nodes."""

    parser: Parser
    macro_handler: MacroHandler

    def __call__(self, stream: TokenStream) -> AstRoot:
        node: AstRoot = self.parser(stream)

        lexical_scope = get_stream_lexical_scope(stream)
        if isinstance(lexical_scope, ClassScope):
            return node

        return self.resolve_deferred(node, stream)

    def resolve_deferred(self, node: AstRoot, stream: TokenStream) -> AstRoot:
        should_replace = False
        commands: List[AstCommand] = []

        for command in node.commands:
            if command.arguments and isinstance(
                body := command.arguments[-1], AstClassRoot
            ):
                resolved_body = self.resolve_deferred(body, stream)
                if resolved_body is not body:
                    should_replace = True
                    command = replace(
                        command,
                        arguments=AstChildren([*command.arguments[:-1], resolved_body]),
                    )

            stack: List[AstCommand] = [command]

            while command.arguments and isinstance(
                command := command.arguments[-1], AstCommand
            ):
                stack.append(command)

            command = stack.pop()

            if command.arguments and isinstance(
                deferred_root := command.arguments[-1], AstDeferredRoot
            ):
                should_replace = True

                self.macro_handler.flush_pending_macros(stream)

                deferred_stream = deferred_root.stream
                deferred_stream.data["loop"] = False
                deferred_stream.data["memo"] = False
                deferred_stream.data["local_spec"] = False
                deferred_stream.data["spec"] = get_stream_spec(stream)
                deferred_stream.data["macro_scope"] = get_stream_macro_scope(stream)
                deferred_stream.data["pending_macros"] = []

                nested_root = delegate("nested_root", deferred_stream)

                self.macro_handler.cache_local_spec(deferred_stream)

                command = replace(
                    command,
                    arguments=AstChildren([*command.arguments[:-1], nested_root]),
                )
                while stack:
                    parent = stack.pop()
                    command = replace(
                        parent,
                        arguments=AstChildren([*parent.arguments[:-1], command]),
                    )

            elif stack:
                command = stack[0]

            commands.append(command)

        if should_replace:
            return replace(node, commands=AstChildren(commands))

        return node


@dataclass
class LexicalScopeConstraint:
    """Constraint that restricts specific statements to the given scope type."""

    parser: Parser
    type: Type[LexicalScope]
    flags: Dict[str, bool]
    command_identifiers: Set[str]

    def __call__(self, stream: TokenStream) -> AstRoot:
        node = self.parser(stream)

        lexical_scope = get_stream_lexical_scope(stream)
        if isinstance(lexical_scope, self.type) and all(
            stream.data.get(flag, False) == enabled
            for flag, enabled in self.flags.items()
        ):
            return node

        if isinstance(node, AstRoot):
            for command in node.commands:
                if command.identifier in self.command_identifiers:
                    name, _, _ = command.identifier.partition(":")
                    exc = InvalidSyntax(
                        f'Statement "{name}" is not allowed in this context.'
                    )
                    raise set_location(exc, command)

        return node


@dataclass
class RootScopeHandler:
    """Handle root scope."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        with stream.provide(root_scope="root_scope" not in stream.data):
            # Resetting the underlying token generator shouldn't be necessary but
            # currently overloading between nested_root/nested_yaml drops tokens
            # for some reason (see summon macro example).
            stream.generator = stream.generate_tokens()
            return self.parser(stream)


@dataclass
class BinaryParser:
    """Parser for binary expressions."""

    operators: List[str]
    parser: Parser
    right_associative: bool = False

    def parse_operands(
        self, stream: TokenStream
    ) -> Tuple[List[AstExpression], List[str]]:
        with stream.syntax(operator="|".join(self.operators)):
            nodes = [self.parser(stream)]
            operations: List[str] = []

            for op in stream.collect("operator"):
                nodes.append(self.parser(stream))
                operations.append(normalize_whitespace(op.value))

        return nodes, operations

    def __call__(self, stream: TokenStream) -> Any:
        nodes, operations = self.parse_operands(stream)

        if self.right_associative:
            result = nodes[-1]
            nodes = nodes[-2::-1]
            operations = operations[::-1]
        else:
            result = nodes[0]
            nodes = nodes[1:]

        for op, node in zip(operations, nodes):
            if self.right_associative:
                result, node = node, result
            result = AstExpressionBinary(operator=op, left=result, right=node)
            result = set_location(result, result.left, result.right)

        return result


@dataclass
class UnaryParser:
    """Parser for unary expressions."""

    operators: List[str]
    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        with stream.syntax(operator="|".join(self.operators)):
            if op := stream.get("operator"):
                operator = normalize_whitespace(op.value)
                node = AstExpressionUnary(operator=operator, value=self(stream))
                return set_location(node, op, node.value)
            return self.parser(stream)


@dataclass
class ChainedComparisonParser(BinaryParser):
    """Parser for chained comparisons."""

    def __call__(self, stream: TokenStream) -> Any:
        nodes, operations = self.parse_operands(stream)

        if len(nodes) == 1:
            return nodes[0]

        if len(operations) == 1:
            node = AstExpressionBinary(
                operator=operations[0], left=nodes[0], right=nodes[1]
            )
            return set_location(node, node.left, node.right)

        node = AstChainedComparison(
            operators=tuple(operations), operands=AstChildren(nodes)
        )
        return set_location(node, nodes[0], nodes[-1])


@dataclass
class UnpackParser:
    """Parser for unpacking."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        with stream.syntax(prefix=r"\*\*|\*"):
            prefix = stream.expect("prefix")

        node = self.parser(stream)

        node = AstUnpack(type="dict" if prefix.value == "**" else "list", value=node)
        return set_location(node, prefix, node.value)


@dataclass
class UnpackConstraint:
    """Constraint for unpacking."""

    type: str
    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        if isinstance(node := self.parser(stream), AstUnpack):
            if node.type != self.type:
                exc = InvalidSyntax(f"{node.type.capitalize()} unpacking not allowed.")
                raise node.emit_error(exc)
        return node


@dataclass
class KeywordParser:
    """Parser for keywords."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        with stream.syntax(name=IDENTIFIER_PATTERN, equal=r"=(?!=)"):
            name = stream.expect("name")
            stream.expect("equal")

        node = self.parser(stream)

        node = AstKeyword(name=name.value, value=node)
        return set_location(node, name, node.value)


@dataclass
class LookupParser:
    """Parser for lookups."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        start = None
        stop = None
        step = None

        with (
            stream.provide(bolt_lookup=True),
            stream.syntax(
                colon=r":",
                comma=r",",
                bracket=r"\]",
            ),
        ):
            colon1 = stream.get("colon")

            if not colon1:
                start = self.parser(stream)
                location = start.location
                colon1 = stream.get("colon")
            else:
                location = colon1.location

            if not colon1:
                return start

            colon2 = stream.get("colon")

            if not colon2:
                with stream.checkpoint():
                    sep = stream.get("comma", "bracket")
                if not sep:
                    stop = self.parser(stream)
                    colon2 = stream.get("colon")

            if colon2:
                with stream.checkpoint():
                    sep = stream.get("comma", "bracket")
                if not sep:
                    step = self.parser(stream)

        node = AstSlice(start=start, stop=stop, step=step)
        return set_location(node, location, stream.current)


@dataclass
class PrimaryParser:
    """Parser for primary expressions."""

    parser: Parser
    truncate: bool = False

    quote_helper: QuoteHelper = field(default_factory=JsonQuoteHelper)

    def __call__(self, stream: TokenStream) -> Any:
        with stream.syntax(
            brace=r"\(|\)",
            comma=r",",
            format_string=r"f['\"]",
            nested_location=r"~/",
        ):
            token = stream.get(("brace", "("), "format_string", "nested_location")

            if token and token.match("brace"):
                with stream.ignore("newline"):
                    comma = None
                    items: List[AstExpression] = []

                    for _ in stream.peek_until(("brace", ")")):
                        items.append(delegate("bolt:expression", stream))

                        if not (comma := stream.get("comma")):
                            stream.expect(("brace", ")"))
                            break

                    if len(items) == 1 and not comma:
                        node = items[0]
                    else:
                        node = AstTuple(items=AstChildren(items))
                        node = set_location(node, token, stream.current)

                if self.truncate:
                    return node

            elif token and token.match("format_string"):
                quote = token.value[-1]

                with (
                    stream.provide(bolt_format_string=True),
                    stream.syntax(
                        escape=rf"\\.",
                        double_brace=r"\{\{|\}\}",
                        brace=r"\{|\}",
                        quote=quote,
                        text=r"[^\\]+?",
                    ),
                ):
                    fmt = quote
                    values: List[AstExpression] = []

                    for escape, double_brace, brace, text in stream.collect(
                        "escape",
                        "double_brace",
                        ("brace", "{"),
                        "text",
                    ):
                        if escape:
                            fmt += escape.value
                        elif double_brace:
                            fmt += double_brace.value
                        elif brace:
                            fmt += "{"
                            with stream.syntax(text=None):
                                values.append(delegate("bolt:expression", stream))
                            with stream.syntax(spec=r"[:!][^\}]+", double_brace=None):
                                if spec := stream.get("spec"):
                                    fmt += spec.value
                                stream.expect(("brace", "}"))
                            fmt += "}"
                        elif text:
                            fmt += text.value

                    end_quote = stream.expect("quote")
                    fmt += end_quote.value

                    fmt = self.quote_helper.unquote_string(
                        Token(
                            "format_string",
                            fmt,
                            token.location.with_horizontal_offset(1),
                            end_quote.end_location,
                        )
                    )

                    node = AstFormatString(fmt=fmt, values=AstChildren(values))
                    node = set_location(node, token, end_quote)

                if self.truncate:
                    return node

            elif token and token.match("nested_location"):
                with (
                    stream.intercept("whitespace"),
                    stream.provide(bolt_nested_location=True),
                    stream.syntax(
                        path=r"[0-9a-z_./-]+",
                        curly=r"\{|\}",
                    ),
                ):
                    fmt = ""
                    values: List[AstExpression] = []

                    for path, curly in stream.collect("path", ("curly", "{")):
                        if path:
                            fmt += path.value
                        elif curly:
                            fmt += "{"
                            with stream.syntax(path=None), stream.ignore("whitespace"):
                                values.append(delegate("bolt:expression", stream))
                            with stream.syntax(spec=r"[:!][^\}]+"):
                                if spec := stream.get("spec"):
                                    fmt += spec.value
                                stream.expect(("curly", "}"))
                            fmt += "}"

                    if not values and self.truncate:
                        exc = InvalidSyntax(
                            "Truncated nested location requires interpolated values."
                        )
                        raise set_location(exc, token, stream.current)

                    node = AstFormattedLocation(fmt=fmt, values=AstChildren(values))
                    node = set_location(node, token, stream.current)

            else:
                node = self.parser(stream)

        with (
            stream.intercept(*["whitespace"] * self.truncate),
            stream.syntax(
                dot=r"\.",
                comma=r",",
                brace=r"\(|\)",
                bracket=r"\[|\]",
                identifier=IDENTIFIER_PATTERN,
                string=STRING_PATTERN,
                number=r"(?:0|[1-9][0-9]*)",
            ),
        ):
            while token := stream.get("dot", ("brace", "("), ("bracket", "[")):
                arguments: List[Any] = []

                if token.match("dot"):
                    identifier, string, number = stream.expect(
                        "identifier",
                        "string",
                        "number",
                    )

                    if identifier:
                        node = AstAttribute(value=node, name=identifier.value)
                        node = set_location(node, node.value, identifier)
                        continue

                    if string:
                        value = (
                            string.value[2:-1]
                            if string.value.startswith("r")
                            else self.quote_helper.unquote_string(string)
                        )
                    elif number:
                        value = int(number.value)

                    arguments.append(set_location(AstValue(value=value), stream.current))  # type: ignore

                else:
                    if token.match("brace"):
                        close = ("brace", ")")
                        argument_parser = delegate("bolt:call_argument")
                    else:
                        close = ("bracket", "]")
                        argument_parser = delegate("bolt:lookup_argument")

                    allow_positional = True

                    with stream.ignore("newline", "whitespace"):
                        for _ in stream.peek_until(close):
                            argument = argument_parser(stream)

                            if isinstance(argument, AstKeyword):
                                allow_positional = False
                            elif isinstance(argument, AstUnpack):
                                if argument.type == "dict":
                                    allow_positional = False
                                elif not allow_positional:
                                    exc = InvalidSyntax(
                                        "List unpacking not allowed after keyword arguments."
                                    )
                                    raise argument.emit_error(exc)
                            elif not allow_positional:
                                exc = InvalidSyntax(
                                    "Positional argument not allowed after keyword arguments."
                                )
                                raise argument.emit_error(exc)

                            arguments.append(argument)

                            if not stream.get("comma"):
                                stream.expect(close)
                                break

                if token.match("brace"):
                    node = AstCall(value=node, arguments=AstChildren(arguments))
                else:
                    if not arguments:
                        arguments = [
                            set_location(AstSlice(), node.end_location, stream.current)
                        ]
                    node = AstLookup(value=node, arguments=AstChildren(arguments))

                node = set_location(node, node.value, stream.current)

        return node


@dataclass
class BuiltinCallRestriction:
    """Only allow call expressions on builtins."""

    parser: Parser
    builtins: Set[str]

    def __call__(self, stream: TokenStream) -> Any:
        parent = None
        node = self.parser(stream)
        original = node

        while isinstance(node, (AstAttribute, AstLookup, AstCall)):
            parent = node
            node = node.value

        lexical_scope = get_stream_lexical_scope(stream)
        if (
            isinstance(node, AstIdentifier)
            and not isinstance(parent, AstCall)
            and node.value in self.builtins  # type: ignore
            and not lexical_scope.has_binding(node.value, search_parents=True)
        ):
            # Reset the underlying token generator so that the identifier can
            # be re-parsed as a different token. That shouldn't be necessary
            # but otherwise this currently induces an error in files that aren't
            # terminated by a newline.
            stream.generator = stream.generate_tokens()
            msg = f'Expected call expression on builtin "{node.value}".'
            raise set_location(InvalidSyntax(msg), parent)

        return original


def parse_dict_item(stream: TokenStream) -> Any:
    """Parse dict item node."""

    with stream.syntax(colon=r":", identifier=IDENTIFIER_PATTERN):
        with stream.checkpoint() as commit:
            token = stream.expect("identifier")
            stream.expect("colon")
            commit()

            lexical_scope = get_stream_lexical_scope(stream)
            if lexical_scope.has_binding(token.value, search_parents=True):
                key = set_location(AstIdentifier(value=token.value), token)
                lexical_scope.reference_variable(key.value, key)
            else:
                key = set_location(AstDictUnquotedKey(value=token.value), token)

        if commit.rollback:
            key = delegate("bolt:expression", stream)
            stream.expect("colon")

        value = delegate("bolt:expression", stream)

    item = AstDictItem(key=key, value=value)
    return set_location(item, key, value)


@dataclass
class LiteralParser:
    """Parser for literals."""

    database: CompilationDatabase
    quote_helper: QuoteHelper = field(default_factory=JsonQuoteHelper)

    def __call__(self, stream: TokenStream) -> Any:
        with stream.syntax(
            curly=r"\{|\}",
            bracket=r"\[|\]",
            comma=r",",
            ellipsis=r"\.\.\.",
            true=TRUE_PATTERN,
            false=FALSE_PATTERN,
            null=NULL_PATTERN,
            docstring=DOCSTRING_PATTERN,
            string=STRING_PATTERN,
            resource=(
                None
                if stream.data.get("bolt_lookup")
                or stream.data.get("bolt_format_string")
                or stream.data.get("bolt_nested_location")
                else RESOURCE_LOCATION_PATTERN
            ),
            number=NUMBER_PATTERN,
        ):
            (
                curly,
                bracket,
                ellipsis,
                true,
                false,
                null,
                docstring,
                string,
                resource,
                number,
            ) = stream.expect(
                ("curly", "{"),
                ("bracket", "["),
                "ellipsis",
                "true",
                "false",
                "null",
                "docstring",
                "string",
                "resource",
                "number",
            )

            if curly:
                items: List[Any] = []
                unquoted = False

                with stream.ignore("newline"):
                    for _ in stream.peek_until(("curly", "}")):
                        items.append(item := delegate("bolt:dict_item", stream))
                        unquoted = unquoted or (
                            isinstance(item, AstDictItem)
                            and isinstance(item.key, AstDictUnquotedKey)
                        )

                        if not stream.get("comma"):
                            stream.expect(("curly", "}"))
                            break

                if unquoted:
                    for i, item in enumerate(items):
                        if isinstance(item, AstDictItem):
                            if isinstance(item.key, AstIdentifier):
                                key = AstDictUnquotedKey(value=item.key.value)
                                key = set_location(key, item.key)
                                items[i] = replace(item, key=key)
                            elif not isinstance(item.key, AstDictUnquotedKey):
                                exc = InvalidSyntax(
                                    "Forbidden dynamic key in dict without quotes."
                                )
                                raise set_location(exc, item.key)

                node = AstDict(items=AstChildren(items))
                return set_location(node, curly, stream.current)

            if bracket:
                elements: List[Any] = []

                with stream.ignore("newline"):
                    for _ in stream.peek_until(("bracket", "]")):
                        elements.append(delegate("bolt:list_item", stream))

                        if not stream.get("comma"):
                            stream.expect(("bracket", "]"))
                            break

                node = AstList(items=AstChildren(elements))
                return set_location(node, bracket, stream.current)

            if ellipsis:
                value = ...
            elif true:
                value = True
            elif false:
                value = False
            elif null:
                value = None
            elif docstring:
                value = (
                    docstring.value[4:-3]
                    if docstring.value.startswith("r")
                    else self.quote_helper.unquote_string(
                        docstring._replace(value=docstring.value[2:-2])
                    )
                )
            elif string:
                value = (
                    string.value[2:-1]
                    if string.value.startswith("r")
                    else self.quote_helper.unquote_string(string)
                )
            elif resource:
                if resource.value.startswith(("./", "../")):
                    value = ":".join(
                        resolve_using_database(
                            relative_path=resource.value,
                            database=self.database,
                            location=resource.location,
                            end_location=resource.end_location,
                        )
                    )
                else:
                    value = resource.value
            elif number:
                value = string_to_number(number.value)

            node = AstValue(value=value)  # type: ignore
            return set_location(node, stream.current)



__all__ = [
    "Codegen",
    "Accumulator",
    "CodegenStatement",
    "WithStatementFusion",
    "ChildrenCollector",
    "CommandCollector",
    "RootCommandCollector",
    "visit_single",
    "visit_multiple",
    "visit_generic",
    "visit_body",
    "visit_binding",
]


from contextlib import ExitStack, contextmanager
from dataclasses import dataclass, field, fields, replace
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Literal,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)

from mecha import (
    AstChildren,
    AstCommand,
    AstNode,
    AstResourceLocation,
    AstRoot,
    Visitor,
    rule,
)

from .ast import (
    AstAssignment,
    AstAttribute,
    AstCall,
    AstChainedComparison,
    AstClassBases,
    AstClassName,
    AstDict,
    AstDictItem,
    AstDocstring,
    AstEscapeRoot,
    AstExpressionBinary,
    AstExpressionUnary,
    AstFormatString,
    AstFormattedLocation,
    AstFromImport,
    AstFunctionSignature,
    AstFunctionSignatureArgument,
    AstFunctionSignaturePositionalMarker,
    AstFunctionSignatureVariadicArgument,
    AstFunctionSignatureVariadicKeywordArgument,
    AstFunctionSignatureVariadicMarker,
    AstIdentifier,
    AstImportedItem,
    AstImportedMacro,
    AstInterpolation,
    AstKeyword,
    AstList,
    AstLookup,
    AstMacro,
    AstMacroCall,
    AstMacroMatchArgument,
    AstMemo,
    AstPrelude,
    AstProcMacro,
    AstSlice,
    AstStatement,
    AstTargetAttribute,
    AstTargetIdentifier,
    AstTargetItem,
    AstTargetUnpack,
    AstTuple,
    AstTypeDeclaration,
    AstUnpack,
    AstValue,
    AstWithContext,
)
from .module import CodegenResult, MacroLibrary


@dataclass(slots=True)
class CodegenStatement:
    """Python statement emitted by the codegen, which can recursively contain other statements."""

    code: Tuple[str, ...]
    lineno: Optional[int] = None
    children: List["CodegenStatement"] = field(default_factory=list)

    def flatten(self, indent: str = "") -> Iterable[Tuple[str, Optional[int]]]:
        """Yield the indented statements with their associated line number."""
        if self.code:
            statement = " ".join(self.code)
            if self.children:
                statement += ":"
            yield f"{indent}{statement}", self.lineno
        if self.children:
            indent += 4 * " "
            for child_statement in self.children:
                yield from child_statement.flatten(indent)


@dataclass
class Accumulator:
    """Utility for generating python code."""

    refs: List[Any] = field(default_factory=list)
    dependencies: Set[str] = field(default_factory=set)
    prelude_imports: List[AstPrelude] = field(default_factory=list)
    macros: MacroLibrary = field(default_factory=dict)
    macro_ids: Dict[str, int] = field(default_factory=dict)
    memo_index: Dict[AstMemo, int] = field(default_factory=dict)
    statements: List["CodegenStatement"] = field(default_factory=list)
    counter: int = 0
    header: Dict[str, str] = field(default_factory=dict)
    root_scope: bool = True

    current_siblings: Tuple[AstNode, ...] = ()
    current_sibling_index: int = 0
    condition_inverse: Optional[str] = None

    def get_source(self) -> str:
        """Return the source code."""
        header = [
            CodegenStatement((variable, "=", expression))
            for variable, expression in self.header.items()
        ]

        lines: List[str] = ["_bolt_lineno = "]
        numbers1: List[int] = [1]
        numbers2: List[int] = [1]

        for statement in header + self.statements:
            for code, lineno in statement.flatten():
                if lineno and numbers2[-1] != lineno:
                    numbers1.append(len(lines) + 1)
                    numbers2.append(lineno)
                lines.append(code)

        lines[0] += f"{numbers1}, {numbers2}"

        return "\n".join(lines) + "\n"

    def helper(self, name: str, *args: Any) -> str:
        """Emit helper."""
        helper = f"_bolt_helper_{name}"

        if helper not in self.header:
            self.header[helper] = f"_bolt_runtime.helpers[{name!r}]"

        return f"{helper}({', '.join(map(str, args))})"

    def replace(self, node: str, **kwargs: str) -> str:
        """Emit replace helper."""
        return self.helper("replace", node, *(f"{k}={v}" for k, v in kwargs.items()))

    def missing(self) -> str:
        """Emit missing sentinel."""
        self.header["_bolt_helper_missing"] = "_bolt_runtime.helpers['missing']"
        return "_bolt_helper_missing"

    def children(self, nodes: Iterable[str]) -> str:
        """Emit children helper."""
        return self.helper("children", f"[{', '.join(nodes)}]")

    def get_attribute_handler(self, obj: str, name: str) -> str:
        """Emit get_attribute_handler helper."""
        attribute_handler = self.helper("get_attribute_handler", obj)
        return f'{attribute_handler}["{name}"]'

    def import_module(self, name: str) -> str:
        """Emit import_module helper."""
        return self.helper("import_module", repr(name))

    def make_ref(self, obj: Any) -> str:
        """Register ref."""
        index = len(self.refs)
        self.refs.append(obj)
        return f"_bolt_refs[{index}]"

    def make_ref_slice(self, objs: Iterable[Any]) -> str:
        """Register ref slice."""
        start = len(self.refs)
        self.refs.extend(objs)
        stop = len(self.refs)
        return f"_bolt_refs[{start}:{stop}]"

    def make_variable(self) -> str:
        """Create a unique variable name."""
        name = f"_bolt_var{self.counter}"
        self.counter += 1
        return name

    def make_macro(self, node: AstMacro) -> str:
        """Add macro to macro library."""
        macro = self.get_macro(node.identifier)
        group = node.identifier.partition(":")[0]
        if self.root_scope:
            self.macros.setdefault(group, {})[macro, node] = None
        return macro

    def import_macro(self, resource_location: str, node: AstImportedMacro) -> str:
        """Import macro into macro library."""
        macro = self.get_macro(node.declaration.identifier)
        group = node.declaration.identifier.partition(":")[0]
        if self.root_scope:
            self.macros.setdefault(group, {})[macro, node.declaration] = (
                resource_location,
                node.name,
            )
        return macro

    def get_macro(self, name: str) -> str:
        """Get macro."""
        if name not in self.macro_ids:
            self.macro_ids[name] = len(self.macro_ids)
        return f"_bolt_macro{self.macro_ids[name]}"

    def extract_lineno(self, lineno: Any):
        """Utility to extract the line number."""
        if isinstance(lineno, AstNode) and not lineno.location.unknown:
            lineno = lineno.location.lineno
        if isinstance(lineno, int):
            return lineno
        return None

    @contextmanager
    def block(self):
        """Wrap statements in an indented block."""
        previous_statements = self.statements
        self.statements = self.statements[-1].children
        try:
            yield
        finally:
            self.statements = previous_statements

    def statement(self, *code: str, lineno: Any = None):
        """Emit statement."""
        self.statements.append(CodegenStatement(code, self.extract_lineno(lineno)))

    @contextmanager
    def function(self, name: str, *args: str, return_type: str = ""):
        """Emit function."""
        return_type_annotation = return_type and f" -> {return_type}"
        self.statement(f"def {name}({', '.join(args)}){return_type_annotation}")
        with self.block():
            previous_root = self.root_scope
            self.root_scope = False
            temp_counter = self.counter
            yield
            self.counter = temp_counter
            self.root_scope = previous_root

    @contextmanager
    def if_statement(
        self,
        condition: str,
        inverse: Optional[str] = None,
        *,
        lineno: Any = None,
    ):
        """Emit if statement."""
        branch = self.helper("branch", condition)
        self.statement("with", branch, "as", "_bolt_condition", lineno=lineno)
        with self.block():
            self.statement("if", "_bolt_condition")
            with self.block():
                yield
        self.condition_inverse = inverse

    @contextmanager
    def else_statement(self, *, lineno: Any = None):
        """Emit else statement."""
        if not self.condition_inverse:
            raise ValueError("Condition inverse unavailable.")
        with self.if_statement(self.condition_inverse, lineno=lineno):
            yield

    def binary(self, left: str, op: str, right: str, *, lineno: Any = None):
        """Emit binary operator."""
        if op in ["in", "not_in"]:
            value = self.helper(f"operator_{op}", left, right)
            self.statement(f"{left} = {value}", lineno=lineno)
        else:
            op = op.replace("_", " ")
            self.statement(f"{left} = {left} {op} {right}", lineno=lineno)

    def dup(self, target: str, *, lineno: Any = None) -> str:
        """Emit __dup__()."""
        dup = self.make_variable()
        value = self.helper("get_dup", target)
        self.statement(f"{dup} = {value}", lineno=lineno)
        self.statement("if", f"{dup} is not None")
        with self.block():
            self.statement(f"{target} = {dup}()")
        return dup

    def rebind(self, target: str, op: str, value: str, *, lineno: Any = None):
        """Emit __rebind__()."""
        rebind = self.helper("get_rebind", target)
        self.statement(f"_bolt_rebind = {rebind}", lineno=lineno)
        self.statement(f"{target} {op} {value}")
        self.statement("if", "_bolt_rebind is not None")
        with self.block():
            self.statement(f"{target} = _bolt_rebind({target})")

    def rebind_dup(self, target: str, dup: str, value: str, *, lineno: Any = None):
        """Emit __rebind__() if target was __dup__()."""
        self.statement("if", f"{dup} is not None")
        with self.block():
            self.rebind(target, "=", value, lineno=lineno)
        self.statement("else")
        with self.block():
            self.statement(f"{target} = {value}")

    def enclose(self, *code: str, from_index: int, lineno: Any = None):
        """Enclose statements starting from the given index."""
        self.statements[from_index:] = [
            CodegenStatement(
                code, self.extract_lineno(lineno), self.statements[from_index:]
            )
        ]


@dataclass
class WithStatementFusion:
    """Finalizer for fusing nested with statements."""

    counter: int = 0

    @classmethod
    def finalize(cls, acc: Accumulator):
        with_statement_fusion = cls()
        acc.statements = [
            with_statement_fusion.fuse(statement, acc) for statement in acc.statements
        ]

    def convert(self, statement: CodegenStatement, exit_stack: str) -> CodegenStatement:
        code = (f"{exit_stack}.enter_context({statement.code[1]})",)
        if len(statement.code) > 2:
            code = (statement.code[3], "=", *code)
        return replace(statement, code=code, children=[])

    def fuse(self, statement: CodegenStatement, acc: Accumulator) -> CodegenStatement:
        children = [self.fuse(child, acc) for child in statement.children]

        if statement.code[0] != "with":
            return replace(statement, children=children)

        nested_children = children
        while nested_children[-1].code[0] == "if":
            nested_children = nested_children[-1].children

        if nested_children[-1].code[0] != "with":
            return replace(statement, children=children)

        nested_statement = nested_children.pop()

        if nested_statement.code[1] == acc.helper("exit_stack"):
            exit_stack = nested_statement.code[3]
            code = nested_statement.code
        else:
            exit_stack = f"_bolt_fused_with_statement{self.counter}"
            self.counter += 1
            code = ("with", acc.helper("exit_stack"), "as", exit_stack)
            nested_children.append(self.convert(nested_statement, exit_stack))

        children.insert(0, self.convert(statement, exit_stack))
        nested_children.extend(nested_statement.children)

        return replace(statement, code=code, children=children)


@dataclass
class ChildrenCollector:
    """Generic children collector."""

    acc: Accumulator
    start_index: int
    children: List[str] = field(default_factory=list)

    def add_static(self, *args: AstNode):
        self.children.extend(self.acc.make_ref(node) for node in args)

    def add_dynamic(self, *args: str):
        self.children.extend(args)

    def flush(self) -> str:
        return self.acc.children(self.children)


class CommandCollector(ChildrenCollector):
    """Collector for commands."""

    def add_static(self, *args: AstNode):
        if len(args) > 1:
            self.acc.statement(
                f"_bolt_runtime.commands.extend({self.acc.make_ref_slice(args)})"
            )
        elif args:
            self.acc.statement(
                f"_bolt_runtime.commands.append({self.acc.make_ref(args[0])})"
            )

    def add_dynamic(self, *args: str):
        for arg in args:
            if arg.startswith("*"):
                self.acc.statement(f"_bolt_runtime.commands.extend({arg[1:]})")
            else:
                self.acc.statement(f"_bolt_runtime.commands.append({arg})")


class RootCommandCollector(CommandCollector):
    """Collector for commands in a standalone root node."""

    def flush(self) -> str:
        commands = self.acc.make_variable()
        self.acc.enclose(
            "with",
            "_bolt_runtime.scope()",
            "as",
            commands,
            from_index=self.start_index,
        )
        return self.acc.helper("children", commands)


@overload
def visit_single(
    node: AstNode,
) -> Generator[AstNode, Optional[List[str]], Optional[str]]: ...


@overload
def visit_single(
    node: AstNode,
    required: Literal[True],
) -> Generator[AstNode, Optional[List[str]], str]: ...


def visit_single(
    node: AstNode,
    required: bool = False,
) -> Generator[AstNode, Optional[List[str]], Optional[str]]:
    """Yield the node and make sure that it returns a single result."""
    result = yield node

    if result is None:
        if required:
            raise ValueError(
                f"Result required for {node.__class__.__name__} {result!r}."
            )
        return None

    if len(result) != 1:
        raise ValueError(
            f"Expected single result for {node.__class__.__name__} {result!r}."
        )

    return result[0]


def visit_multiple(
    children: Tuple[AstNode, ...],
    acc: Accumulator,
    children_collector: Type[ChildrenCollector] = ChildrenCollector,
) -> Generator[AstNode, Optional[List[str]], Optional[str]]:
    """Yield all the nodes and return a single result pointing to the new children."""
    current_count = 0
    collector: Optional[ChildrenCollector] = None
    index = len(acc.statements)

    previous_siblings = acc.current_siblings
    previous_sibling_index = acc.current_sibling_index
    acc.current_siblings = children

    for i, child in enumerate(children):
        acc.current_sibling_index = i

        result = yield child
        if result is None:
            continue
        if not collector:
            collector = children_collector(acc, index)

        statements = acc.statements[index:]
        del acc.statements[index:]
        collector.add_static(*children[current_count:i])
        acc.statements.extend(statements)
        collector.add_dynamic(*result)

        current_count = i + 1
        index = len(acc.statements)

    acc.current_siblings = previous_siblings
    acc.current_sibling_index = previous_sibling_index

    if collector:
        collector.add_static(*children[current_count:])
        return collector.flush()

    return None


def visit_generic(
    node: AstNode,
    acc: Accumulator,
    children_collector: Type[ChildrenCollector] = ChildrenCollector,
) -> Generator[AstNode, Optional[List[str]], Optional[str]]:
    """Recursively yield all the fields and return a result pointing to the updated node."""
    to_replace: Dict[str, str] = {}

    for f in fields(node):
        attribute = getattr(node, f.name)
        result = None

        if isinstance(attribute, AstChildren):
            attribute = cast(AstChildren[AstNode], attribute)
            result = yield from visit_multiple(attribute, acc, children_collector)

        elif isinstance(attribute, AstNode):
            result = yield from visit_single(attribute)

        if result is not None:
            to_replace[f.name] = result

    if not to_replace:
        return None

    return acc.replace(acc.make_ref(node), **to_replace)


def visit_body(
    node: AstRoot,
    acc: Accumulator,
) -> Generator[AstNode, Optional[List[str]], None]:
    """Emit each individual command in the current scope."""
    result = yield from visit_multiple(node.commands, acc, CommandCollector)
    if result is None:
        acc.statement(f"_bolt_runtime.commands.extend({acc.make_ref(node)}.commands)")


def visit_binding(
    target_node: AstNode,
    op: str,
    value: str,
    acc: Accumulator,
) -> Generator[AstNode, Optional[List[str]], None]:
    """Emit variable binding."""
    targets = yield target_node

    if not targets:
        return

    if len(targets) > 1 and isinstance(target_node, AstTargetUnpack):
        nodes = target_node.targets
        values = [f"_bolt_unpack{i}" for i in range(len(targets))]
        acc.statement(f"{', '.join(values)} = {value}", lineno=target_node)
    else:
        nodes = [target_node]
        values = [value]

    for node, target, value in zip(nodes, targets, values):
        if isinstance(node, AstTargetIdentifier) and node.rebind:
            acc.rebind(target, op, value, lineno=node)
        else:
            acc.statement(f"{target} {op} {value}", lineno=node)


@dataclass(eq=False)
class Codegen(Visitor):
    """Code generator."""

    accumulator_factory: Callable[[], Accumulator] = Accumulator
    finalizers: List[Callable[[Accumulator], None]] = field(
        default_factory=lambda: [WithStatementFusion.finalize]
    )

    def __call__(self, node: AstRoot) -> CodegenResult:  # type: ignore
        acc = self.accumulator_factory()
        result = self.invoke(node, acc)

        if result is None:
            return CodegenResult()

        if len(result) != 1:
            raise ValueError(
                f"Expected single result for {node.__class__.__name__} {result!r}."
            )

        output = acc.make_variable()
        acc.statement(f"{output} = {result[0]}")

        for finalizer in self.finalizers:
            finalizer(acc)

        return CodegenResult(
            source=acc.get_source(),
            output=output,
            refs=acc.refs,
            dependencies=acc.dependencies,
            prelude_imports=acc.prelude_imports,
            macros=acc.macros,
        )

    @rule(AstNode)
    def fallback(
        self,
        node: AstNode,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        result = yield from visit_generic(node, acc)
        if result is None:
            return None
        return [result]

    @rule(AstRoot)
    def root(
        self,
        node: AstRoot,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        result = yield from visit_generic(node, acc, RootCommandCollector)
        if result is None:
            return None
        return [result]

    @rule(AstCommand)
    def command(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        if not node.arguments:
            return None

        if not isinstance(node.arguments[-1], (AstCommand, AstRoot)):
            arguments = yield from visit_multiple(node.arguments, acc)
            if arguments is None:
                return None
            return [acc.replace(acc.make_ref(node), arguments=arguments)]

        arguments = yield from visit_multiple(node.arguments[:-1], acc)
        nesting_index = len(acc.statements)
        nesting = yield from visit_single(node.arguments[-1])

        if nesting is None:
            if arguments is None:
                return None
            nesting = acc.make_ref(node.arguments[-1])
        else:
            if arguments is None and len(node.arguments) > 1:
                arguments = acc.make_ref_slice(node.arguments[:-1])
            push_arguments = f", *{arguments}" if arguments else ""
            acc.enclose(
                "with",
                f"_bolt_runtime.push_nesting({node.identifier!r}{push_arguments})",
                from_index=nesting_index,
            )

        arguments = acc.children([f"*{arguments}", nesting] if arguments else [nesting])
        return [acc.replace(acc.make_ref(node), arguments=arguments)]

    @rule(AstStatement)
    def statement(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        yield node.arguments[0]
        return []

    @rule(AstMemo)
    def memo(
        self,
        node: AstMemo,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        file_index = acc.memo_index.setdefault(node, 0)
        acc.memo_index[node] = file_index + 1

        keys: List[str] = []

        *arguments, body = node.arguments
        cached_identifiers = (
            "".join(f"{name}," for name in body.identifiers)
            if isinstance(body, AstEscapeRoot)
            else None
        )

        for arg in arguments:
            if isinstance(arg, AstAssignment):
                target = cast(AstTargetIdentifier, arg.target)
                yield arg
                keys.append(f"{target.value},")
            else:
                value = yield from visit_single(arg, required=True)
                keys.append(f"{value},")

        storage = f"_bolt_memo_storage_{node.persistent_id.hex}"
        acc.header[storage] = "None"
        if not acc.root_scope:
            acc.statement(f"global {storage}", lineno=node)
        acc.statement("if", f"{storage} is None")
        with acc.block():
            acc.statement(
                f"{storage} = _bolt_runtime.memo.registry[__file__][{acc.make_ref(node)}, {file_index}]"
            )

        path = f"_bolt_memo_invocation_path_{node.persistent_id.hex}"
        acc.statement(f"{path} = _bolt_runtime.get_nested_location()")

        invocation = f"_bolt_memo_invocation_{node.persistent_id.hex}"
        acc.statement(f"{invocation} = {storage}[({path}, {' '.join(keys)})]")

        acc.statement("if", f"{invocation}.cached")
        with acc.block():
            acc.statement(f"_bolt_runtime.memo.restore(_bolt_runtime, {invocation})")
            if cached_identifiers:
                acc.statement(f"({cached_identifiers}) = {invocation}.bindings")

        acc.statement("else")
        with acc.block():
            acc.statement(
                "with",
                f"_bolt_runtime.memo.record(_bolt_runtime, {invocation}, {path}, __name__)",
            )
            with acc.block():
                yield from visit_body(cast(AstRoot, body), acc)
            if cached_identifiers:
                acc.statement(f"{invocation}.bindings = ({cached_identifiers})")

        return []

    @rule(AstCommand, identifier="def:function:body")
    def function(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        signature = cast(AstFunctionSignature, node.arguments[0])
        body = cast(AstRoot, node.arguments[1])

        decorators: List[str] = []
        for decorator in signature.decorators:
            value = yield from visit_single(decorator.expression, required=True)
            decorators.append(value)

        arguments: List[str] = []
        for arg in signature.arguments:
            if isinstance(arg, AstFunctionSignatureArgument):
                a = arg.name
                if arg.type_annotation:
                    value = yield from visit_single(arg.type_annotation, required=True)
                    a = f"{a}: {value}"
                if arg.default:
                    padding = " " * bool(arg.type_annotation)
                    a = f"{a}{padding}={padding}{acc.missing()}"
                arguments.append(a)
                f"{arg.name}={acc.missing()}" if arg.default else arg.name
            elif isinstance(arg, AstFunctionSignaturePositionalMarker):
                arguments.append("/")
            elif isinstance(arg, AstFunctionSignatureVariadicArgument):
                a = f"*{arg.name}"
                if arg.type_annotation:
                    value = yield from visit_single(arg.type_annotation, required=True)
                    a = f"{a}: {value}"
                arguments.append(a)
            elif isinstance(arg, AstFunctionSignatureVariadicMarker):
                arguments.append("*")
            elif isinstance(arg, AstFunctionSignatureVariadicKeywordArgument):
                a = f"**{arg.name}"
                if arg.type_annotation:
                    value = yield from visit_single(arg.type_annotation, required=True)
                    a = f"{a}: {value}"
                arguments.append(a)

        return_type = ""
        if signature.return_type_annotation:
            return_type = yield from visit_single(
                signature.return_type_annotation, required=True
            )

        with acc.function(signature.name, *arguments, return_type=return_type):
            if body.commands and isinstance(body.commands[0], AstDocstring):
                yield body.commands[0]
                body = replace(body, commands=AstChildren(body.commands[1:]))

            for arg in signature.arguments:
                if isinstance(arg, AstFunctionSignatureArgument) and arg.default:
                    acc.statement("if", f"{arg.name} is {acc.missing()}")
                    with acc.block():
                        value = yield from visit_single(arg.default, required=True)
                        acc.statement(f"{arg.name} = {value}")

            yield from visit_body(body, acc)

        for decorator, value in list(zip(signature.decorators, decorators))[::-1]:
            acc.statement(
                f"{signature.name} = {value}({signature.name})",
                lineno=decorator,
            )

        return []

    @rule(AstCommand, identifier="return")
    @rule(AstCommand, identifier="return:pythonresult")
    def return_statement(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        statement = "return"
        if node.arguments:
            value = yield from visit_single(node.arguments[0], required=True)
            statement += f" {value}"
        acc.statement(statement)
        return []

    @rule(AstCommand, identifier="yield")
    @rule(AstCommand, identifier="yield:value")
    @rule(AstCommand, identifier="yield:from:value")
    def yield_statement(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        statement = "yield from" if node.identifier == "yield:from:value" else "yield"
        if node.arguments:
            value = yield from visit_single(node.arguments[0], required=True)
            statement += f" {value}"
        acc.statement(statement)
        return []

    @rule(AstMacro)
    def macro(
        self,
        node: AstMacro,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        macro = acc.make_macro(
            replace(node, arguments=AstChildren(node.arguments[:-1]))
        )
        arguments = [
            arg.match_identifier.value
            for arg in node.arguments
            if isinstance(arg, AstMacroMatchArgument)
        ]

        with acc.function(macro, *arguments):
            yield from visit_body(cast(AstRoot, node.arguments[-1]), acc)

        return []

    @rule(AstMacroCall)
    def macro_call(
        self,
        node: AstMacroCall,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        macro = acc.get_macro(node.identifier)
        result = yield from visit_generic(node, acc)
        if result is None:
            result = acc.make_ref(node)

        macro_call = acc.helper("macro_call", "_bolt_runtime", macro, result)
        result = acc.make_variable()
        acc.statement(f"{result} = {macro_call}", lineno=node)

        return [f"*{result}"]

    @rule(AstProcMacro)
    def proc_macro(
        self,
        node: AstProcMacro,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        macro = acc.make_macro(
            replace(node, arguments=AstChildren(node.arguments[:-1]))
        )

        with acc.function(macro, "stream"):
            yield from visit_body(cast(AstRoot, node.arguments[-1]), acc)

        if acc.root_scope:
            acc.header["_bolt_proc_macros"] = "{}"
            acc.statement(f"_bolt_proc_macros[{node.identifier!r}] = {macro}")

        return []

    @rule(AstCommand, identifier="class:name:bases:body")
    @rule(AstCommand, identifier="class:name:body")
    def class_definition(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        name = cast(AstClassName, node.arguments[0])
        body = cast(AstRoot, node.arguments[-1])

        decorators: List[str] = []
        for decorator in name.decorators:
            value = yield from visit_single(decorator.expression, required=True)
            decorators.append(value)

        class_args: List[str] = []

        if isinstance(bases := node.arguments[1], AstClassBases):
            for base in bases.inherit:
                result = yield from visit_single(base, required=True)
                class_args.append(result)
            for kwarg in bases.kwargs:
                result = yield from visit_single(kwarg, required=True)
                class_args.append(result)

        joined_args = f"({', '.join(class_args)})" if class_args else ""
        acc.statement(f"class {name.value}{joined_args}", lineno=node)

        with acc.block():
            temp_start = acc.counter
            yield from visit_body(body, acc)
            temp_stop = acc.counter

            for i in range(temp_start, temp_stop):
                acc.statement(f"del _bolt_var{i}")

            acc.counter = temp_start

        for decorator, value in list(zip(name.decorators, decorators))[::-1]:
            acc.statement(
                f"{name.value} = {value}({name.value})",
                lineno=decorator,
            )

        return []

    @rule(AstDocstring)
    def docstring(
        self,
        node: AstDocstring,
        acc: Accumulator,
    ) -> Optional[List[str]]:
        acc.statement(repr(cast(AstValue, node.arguments[0]).value))
        return []

    @rule(AstCommand, identifier="del:target")
    def del_statement(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        targets = yield node.arguments[0]
        for target in targets or []:
            acc.statement(f"del {target}")
        return []

    @rule(AstCommand, identifier="if:condition:body")
    def if_statement(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        condition = yield from visit_single(node.arguments[0], required=True)
        inverse = None

        else_index = acc.current_sibling_index + 1
        if (
            else_index < len(acc.current_siblings)
            and isinstance(else_node := acc.current_siblings[else_index], AstCommand)
            and else_node.identifier == "else:body"
        ):
            inverse = f"{condition}_inverse"
            value = acc.helper("operator_not", condition)
            acc.statement(f"{inverse} = {value}", lineno=node.arguments[0])

        with acc.if_statement(condition, inverse, lineno=node):
            yield from visit_body(cast(AstRoot, node.arguments[1]), acc)
        return []

    @rule(AstCommand, identifier="else:body")
    def else_statement(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        with acc.else_statement(lineno=node):
            yield from visit_body(cast(AstRoot, node.arguments[0]), acc)
        return []

    @rule(AstCommand, identifier="while:condition:body")
    def while_statement(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        acc.statement(f"while True", lineno=node)
        with acc.block():
            acc.statement(
                "with", "_bolt_runtime.scope()", "as", "_bolt_condition_commands"
            )
            with acc.block():
                condition = yield from visit_single(node.arguments[0], required=True)

            again = acc.make_variable()
            loop = acc.helper("loop", condition)
            acc.statement(
                "with",
                loop,
                "as",
                f"(_bolt_loop_overridden, {again})",
                lineno=node.arguments[0],
            )
            with acc.block():
                acc.statement("_bolt_runtime.commands.extend(_bolt_condition_commands)")

                acc.statement(
                    "if", "not _bolt_loop_overridden", lineno=node.arguments[0]
                )
                with acc.block():
                    acc.statement(f"{condition} = bool({condition})")

                with acc.if_statement(condition, lineno=node.arguments[0]):
                    yield from visit_body(cast(AstRoot, node.arguments[1]), acc)

                    acc.statement("if", again, lineno=node.arguments[0])
                    with acc.block():
                        acc.statement("continue")

            acc.statement("break")

        return []

    @rule(AstCommand, identifier="for:target:in:iterable:body")
    def for_statement(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        iterable = yield from visit_single(node.arguments[1], required=True)
        targets = yield node.arguments[0]
        acc.statement(f"for {', '.join(targets or ['_'])} in {iterable}")
        with acc.block():
            yield from visit_body(cast(AstRoot, node.arguments[2]), acc)
        return []

    @rule(AstCommand, identifier="break")
    def break_statement(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Optional[List[str]]:
        acc.statement("break")
        return []

    @rule(AstCommand, identifier="continue")
    def continue_statement(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Optional[List[str]]:
        acc.statement("continue")
        return []

    @rule(AstCommand, identifier="with:context:body")
    def with_statement(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        context = cast(AstWithContext, node.arguments[0])
        body = cast(AstRoot, node.arguments[1])

        with ExitStack() as exit_stack:
            for clause in context.clauses:
                value = yield from visit_single(clause.value, required=True)

                if clause.target:
                    acc.statement("with", value, "as", "_bolt_with", lineno=clause)
                    exit_stack.enter_context(acc.block())
                    yield from visit_binding(clause.target, "=", "_bolt_with", acc)
                else:
                    acc.statement("with", value, lineno=clause)
                    exit_stack.enter_context(acc.block())

            yield from visit_body(body, acc)

        return []

    @rule(AstCommand, identifier="import:module")
    @rule(AstCommand, identifier="import:module:as:alias")
    def import_statement(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Optional[List[str]]:
        module = cast(AstResourceLocation, node.arguments[0])

        if node.identifier == "import:module:as:alias":
            alias = cast(AstImportedItem, node.arguments[1]).name

            if module.namespace:
                full_path = module.get_value()
                acc.statement(
                    f"{alias} = _bolt_runtime.import_module({full_path!r}).namespace",
                    lineno=node,
                )
                acc.dependencies.add(full_path)
            else:
                rhs = acc.import_module(module.path)
                acc.statement(f"{alias} = {rhs}", lineno=node)

        else:
            name = module.path.partition(".")[0]
            rhs = acc.import_module(module.path)
            if name == module.path:
                acc.statement(f"{name} = {rhs}", lineno=node)
            else:
                acc.statement(rhs, lineno=node)
                acc.statement(f"{name} = {acc.import_module(name)}")

        return []

    @rule(AstFromImport)
    def from_import_statement(
        self,
        node: AstFromImport,
        acc: Accumulator,
    ) -> Optional[List[str]]:
        if isinstance(node, AstPrelude):
            acc.prelude_imports.append(node)

        module = cast(AstResourceLocation, node.arguments[0])
        items = cast(
            AstChildren[Union[AstImportedItem, AstImportedMacro]], node.arguments[1:]
        )
        names = [item.name for item in items]

        if module.namespace:
            full_path = module.get_value()
            targets: List[str] = []
            for item in items:
                if isinstance(item, AstImportedMacro):
                    targets.append(acc.import_macro(full_path, item))
                else:
                    targets.append(item.name)
            stmt = f"{', '.join(targets)} = _bolt_runtime.from_module_import({full_path!r}, {', '.join(map(repr, names))})"
            acc.statement(stmt, lineno=node)
            acc.dependencies.add(full_path)
        else:
            stmt = f"_bolt_from_import = {acc.import_module(module.path)}"
            acc.statement(stmt, lineno=node)
            for name in names:
                rhs = acc.get_attribute_handler("_bolt_from_import", name)
                acc.statement(f"{name} = {rhs}", lineno=node)

        return []

    @rule(AstCommand, identifier="global:subcommand")
    @rule(AstCommand, identifier="nonlocal:subcommand")
    def global_nonlocal_statement(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Optional[List[str]]:
        storage, _, _ = node.identifier.partition(":")

        names: List[str] = []
        subcommand = cast(AstCommand, node.arguments[0])

        while True:
            if isinstance(name := subcommand.arguments[0], AstIdentifier):
                names.append(name.value)
            if subcommand.identifier == f"{storage}:name:subcommand":
                subcommand = cast(AstCommand, subcommand.arguments[1])
            else:
                break

        acc.statement(f"{storage} {', '.join(names)}", lineno=node)

        return []

    @rule(AstInterpolation)
    def interpolation(
        self,
        node: AstInterpolation,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        result = yield from visit_single(node.value, required=True)
        value = "f" + repr(node.prefix + "{" + result + "}") if node.prefix else result

        if node.unpack == "**":
            value = f"{{**{value}}}"
        elif node.unpack == "*":
            value = f"[*{value}]"

        rhs = acc.helper(f"interpolate_{node.converter}", value, acc.make_ref(node))
        acc.statement(f"{result} = {rhs}", lineno=node)

        if node.unpack:
            result = f"*{result}"

        return [result]

    @rule(AstExpressionBinary)
    def binary(
        self,
        node: AstExpressionBinary,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        left = yield from visit_single(node.left, required=True)
        right = yield from visit_single(node.right, required=True)
        acc.binary(left, node.operator, right, lineno=node)
        return [left]

    @rule(AstExpressionBinary, operator="and")
    @rule(AstExpressionBinary, operator="or")
    def binary_logical(
        self,
        node: AstExpressionBinary,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        left = yield from visit_single(node.left, required=True)

        condition = acc.make_variable()
        value = acc.helper("operator_not", left) if node.operator == "or" else left
        acc.statement(f"{condition} = {value}", lineno=node.left)

        dup = acc.dup(left, lineno=node)

        with acc.if_statement(condition, lineno=node):
            right = yield from visit_single(node.right, required=True)
            acc.rebind_dup(left, dup, right, lineno=node.right)

        return [left]

    @rule(AstExpressionUnary)
    def unary(
        self,
        node: AstExpressionUnary,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        result = yield from visit_single(node.value, required=True)
        if node.operator in ["not"]:
            value = acc.helper(f"operator_{node.operator}", result)
            acc.statement(f"{result} = {value}", lineno=node)
        else:
            op = node.operator.replace("_", " ")
            acc.statement(f"{result} = {op} {result}", lineno=node)
        return [result]

    @rule(AstChainedComparison)
    def chained_comparison(
        self,
        node: AstChainedComparison,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        left = yield from visit_single(node.operands[0], required=True)
        right = yield from visit_single(node.operands[1], required=True)
        acc.binary(left, node.operators[0], right, lineno=node)

        condition = acc.make_variable()

        for op, operand in zip(node.operators[1:], node.operands[2:]):
            acc.statement(f"{condition} = {left}")

            dup = acc.dup(left, lineno=node)

            with acc.if_statement(condition, lineno=node):
                current = right
                right = yield from visit_single(operand, required=True)
                acc.binary(current, op, right, lineno=node)
                acc.rebind_dup(left, dup, current, lineno=operand)

        return [left]

    @rule(AstValue)
    def value(self, node: AstValue, acc: Accumulator) -> Optional[List[str]]:
        result = acc.make_variable()
        acc.statement(f"{result} = {node.literal}")
        return [result]

    @rule(AstIdentifier)
    def identifier(self, node: AstIdentifier, acc: Accumulator) -> Optional[List[str]]:
        result = acc.make_variable()
        acc.statement(f"{result} = {node.value}", lineno=node)
        return [result]

    @rule(AstFormatString)
    def format_string(
        self,
        node: AstFormatString,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        values: List[str] = []

        for value in node.values:
            result = yield from visit_single(value, required=True)
            values.append(result)

        result = acc.make_variable()
        acc.statement(
            f"{result} = {node.fmt!r}.format({', '.join(values)})", lineno=node
        )
        return [result]

    @rule(AstFormattedLocation)
    def formatted_location(
        self,
        node: AstFormattedLocation,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        values: List[str] = []

        for value in node.values:
            result = yield from visit_single(value, required=True)
            values.append(result)

        result = acc.make_variable()
        resolved = acc.helper(
            "resolve_formatted_location",
            "_bolt_runtime",
            f"{node.fmt!r}.format({', '.join(values)})" if values else repr(node.fmt),
        )
        acc.statement(f"{result} = {resolved}", lineno=node)
        return [result]

    @rule(AstTuple)
    def tuple(
        self,
        node: AstTuple,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        items: List[str] = []

        for item in node.items:
            value = yield from visit_single(item, required=True)
            items.append(f"{value},")

        result = acc.make_variable()
        acc.statement(f"{result} = ({''.join(items)})", lineno=node)
        return [result]

    @rule(AstList)
    def list(
        self,
        node: AstList,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        items: List[str] = []

        for item in node.items:
            value = yield from visit_single(item, required=True)
            items.append(value)

        result = acc.make_variable()
        acc.statement(f"{result} = [{', '.join(items)}]", lineno=node)
        return [result]

    @rule(AstDict)
    def dict(
        self,
        node: AstDict,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        items: List[str] = []

        for item in node.items:
            value = yield from visit_single(item, required=True)
            items.append(value)

        result = acc.make_variable()
        acc.statement(f"{result} = {{{', '.join(items)}}}", lineno=node)
        return [result]

    @rule(AstDictItem)
    def dict_item(
        self,
        node: AstDictItem,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        key = yield from visit_single(node.key, required=True)
        value = yield from visit_single(node.value, required=True)
        return [f"{key}: {value}"]

    @rule(AstSlice)
    def slice(
        self,
        node: AstSlice,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        start = ""
        stop = ""
        step = ""

        if node.start:
            start = yield from visit_single(node.start, required=True)
        if node.stop:
            stop = yield from visit_single(node.stop, required=True)
        if node.step:
            step = yield from visit_single(node.step, required=True)

        result = f"{start}:{stop}"
        if step:
            result += f":{step}"

        return [result]

    @rule(AstUnpack)
    def unpack(
        self,
        node: AstUnpack,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        value = yield from visit_single(node.value, required=True)
        prefix = "**" if node.type == "dict" else "*"
        return [f"{prefix}{value}"]

    @rule(AstKeyword)
    def keyword(
        self,
        node: AstKeyword,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        value = yield from visit_single(node.value, required=True)
        return [f"{node.name}={value}"]

    @rule(AstAttribute)
    def attribute(
        self,
        node: AstAttribute,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        value = yield from visit_single(node.value, required=True)
        rhs = acc.get_attribute_handler(value, node.name)
        acc.statement(f"{value} = {rhs}", lineno=node)
        return [value]

    @rule(AstLookup)
    def lookup(
        self,
        node: AstLookup,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        result = yield from visit_single(node.value, required=True)
        arguments: List[str] = []

        for arg in node.arguments:
            value = yield from visit_single(arg, required=True)
            arguments.append(value)

        acc.statement(f"{result} = {result}[{', '.join(arguments)}]", lineno=node)
        return [result]

    @rule(AstCall)
    def call(
        self,
        node: AstCall,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        result = yield from visit_single(node.value, required=True)
        arguments: List[str] = []

        for arg in node.arguments:
            value = yield from visit_single(arg, required=True)
            arguments.append(value)

        acc.statement(f"{result} = {result}({', '.join(arguments)})", lineno=node)
        return [result]

    @rule(AstAssignment)
    def assignment(
        self,
        node: AstAssignment,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        value = yield from visit_single(node.value, required=True)
        yield from visit_binding(node.target, node.operator, value, acc)
        if node.type_annotation and isinstance(node.target, AstTargetIdentifier):
            value = yield from visit_single(node.type_annotation, required=True)
            acc.statement(f"{node.target.value}: {value}")
        return []

    @rule(AstTypeDeclaration)
    def type_declaration(
        self,
        node: AstTypeDeclaration,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        value = yield from visit_single(node.type_annotation, required=True)
        acc.statement(f"{node.identifier.value}: {value}")
        return []

    @rule(AstTargetIdentifier)
    def target_identifier(
        self,
        node: AstTargetIdentifier,
        acc: Accumulator,
    ) -> Optional[List[str]]:
        return [node.value]

    @rule(AstTargetUnpack)
    def target_destructure(
        self,
        node: AstTargetUnpack,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        targets: List[str] = []
        for target in node.targets:
            result = yield from visit_single(target, required=True)
            targets.append(result)
        return targets

    @rule(AstTargetAttribute)
    def target_attribute(
        self,
        node: AstTargetAttribute,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        value = yield from visit_single(node.value, required=True)
        target = acc.get_attribute_handler(value, node.name)
        return [target]

    @rule(AstTargetItem)
    def target_item(
        self,
        node: AstTargetItem,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        result = yield from visit_single(node.value, required=True)
        arguments: List[str] = []

        for arg in node.arguments:
            value = yield from visit_single(arg, required=True)
            arguments.append(value)

        return [f"{result}[{', '.join(arguments)}]"]

    @rule(AstCommand, identifier="pass")
    def pass_statement(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Optional[List[str]]:
        acc.statement("pass")
        return []

    @rule(AstCommand, identifier="raise")
    def raise_statement(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Optional[List[str]]:
        acc.statement("raise")
        return []

    @rule(AstCommand, identifier="raise:exception")
    def raise_exc_statement(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        result = yield from visit_single(node.arguments[0], required=True)
        acc.statement(f"raise {result}")
        return []

    @rule(AstCommand, identifier="raise:exception:from:cause")
    def raise_exc_from_statement(
        self,
        node: AstCommand,
        acc: Accumulator,
    ) -> Generator[AstNode, Optional[List[str]], Optional[List[str]]]:
        result = yield from visit_single(node.arguments[0], required=True)
        cause = yield from visit_single(node.arguments[1], required=True)
        acc.statement(f"raise {result} from {cause}")
        return []



__all__ = [
    "LexicalScope",
    "GlobalScope",
    "FunctionScope",
    "MacroScope",
    "ProcMacroScope",
    "ClassScope",
    "Variable",
    "Binding",
    "UndefinedIdentifier",
    "UnboundLocalIdentifier",
    "InconsistentIdentifierStorage",
]


from dataclasses import dataclass, field, replace
from typing import Dict, List, Literal, Optional, Set, Tuple, Type, TypeVar

from beet.core.utils import extra_field
from mecha import AstNode
from tokenstream import InvalidSyntax, set_location

from .utils import suggest_typo

LexicalScopeType = TypeVar("LexicalScopeType", bound="LexicalScope")


class UndefinedIdentifier(InvalidSyntax):
    """Raised when an identifier is not defined.

    Attributes
    ----------
    identifier
        The identifier that couldn't be found.
    lexical_scope
        The current scope.
    """

    identifier: str
    lexical_scope: "LexicalScope"

    def __init__(self, identifier: str, lexical_scope: "LexicalScope"):
        super().__init__(identifier, lexical_scope)
        self.identifier = identifier
        self.lexical_scope = lexical_scope

    def __str__(self) -> str:
        msg = f'Identifier "{self.identifier}" is not defined.'

        variable_names = self.lexical_scope.list_variables() - {self.identifier}
        if suggestion := suggest_typo(self.identifier, variable_names):
            msg += f" Did you mean {suggestion}?"

        return msg


class UnboundLocalIdentifier(UndefinedIdentifier):
    """Raised when mutating a variable before assignment.

    This is a specialized error for code that tries to access and mutate a variable
    with no local binding in the current scope.
    """

    def __init__(self, identifier: str, lexical_scope: "LexicalScope"):
        super().__init__(identifier, lexical_scope)
        if lexical_scope.parent and lexical_scope.parent.has_variable(identifier):
            self.notes.append(
                f'Use "global {identifier}" or "nonlocal {identifier}" to mutate the variable defined in outer scope.'
            )


class InconsistentIdentifierStorage(InvalidSyntax):
    """Raised when trying to change the storage of an already existing variable.

    Attributes
    ----------
    identifier
        The identifier with a prior conflicting definition in the current scope.
    storage
        The target storage of the inconsistent binding.
    lexical_scope
        The current scope.
    """

    identifier: str
    storage: Literal["local", "nonlocal", "global"]
    lexical_scope: "LexicalScope"

    def __init__(
        self,
        identifier: str,
        storage: Literal["local", "nonlocal", "global"],
        lexical_scope: "LexicalScope",
    ):
        super().__init__(identifier, storage, lexical_scope)
        self.identifier = identifier
        self.storage = storage
        self.lexical_scope = lexical_scope

    def __str__(self) -> str:
        return f'Can\'t make {self.lexical_scope.variables[self.identifier].storage} identifier "{self.identifier}" {self.storage}.'


@dataclass
class Binding:
    """Variable binding.

    Attributes
    ----------
    origin
        The code fragment performing the variable binding.
    references
        Other parts of the code that use the value set by the binding.
    cutoff
        The number of references in common after forking.
    """

    origin: AstNode
    references: List[AstNode] = field(default_factory=list)

    cutoff: int = extra_field(default=0)

    def fork(self) -> "Binding":
        return replace(
            self,
            references=self.references.copy(),
            cutoff=len(self.references),
        )

    def reconcile(self, binding: "Binding"):
        self.references += binding.references[binding.cutoff :]


@dataclass
class Variable:
    """Variable definition.

    Attributes
    ----------
    storage
        The type of storage for the variable.
    bindings
        Associated bindings updating the value or the storage of the variable.
    cutoff
        The number of bindings in common after forking.
    """

    storage: Literal["local", "nonlocal", "global", "deleted"] = "local"
    bindings: List[Binding] = field(default_factory=list)

    cutoff: int = extra_field(default=0)

    def fork(self):
        return replace(
            self,
            bindings=[binding.fork() for binding in self.bindings],
            cutoff=len(self.bindings),
        )

    def reconcile(self, variable: "Variable"):
        for i, binding in enumerate(variable.bindings[: variable.cutoff]):
            self.bindings[i].reconcile(binding)

        for new_binding in variable.bindings[variable.cutoff :]:
            self.bindings.append(new_binding)


@dataclass
class LexicalScope:
    """Class for managing variable scopes during parsing.

    Attributes
    ----------
    parent
        Parent scope if the current scope has one.
    variables
        Dictionary holding variable definitions.
    dirty
        Flag to skip reconciliation after forking in case no additional
        variables or bindings were introduced.
    """

    variables: Dict[str, Variable] = field(default_factory=dict)

    anchor: Optional[AstNode] = extra_field(default=None)
    parent: Optional["LexicalScope"] = extra_field(default=None)
    children: List["LexicalScope"] = extra_field(default_factory=list)

    dirty: bool = extra_field(default=False)

    next_branch: Optional["LexicalScope"] = extra_field(default=None)
    next_deferred: Optional["LexicalScope"] = extra_field(default=None)
    pending_bindings: List[Tuple[str, AstNode]] = extra_field(default_factory=list)

    def has_binding(self, identifier: str, search_parents: bool = False) -> bool:
        variable = self.variables.get(identifier)
        return (variable and variable.storage != "deleted") or bool(
            search_parents
            and self.parent
            and self.parent.has_binding(identifier, search_parents=True)
        )

    def reference_binding(self, identifier: str, node: AstNode):
        if not self.has_binding(identifier):
            raise set_location(UnboundLocalIdentifier(identifier, self), node)
        self.reference_variable(identifier, node)

    def has_variable(self, identifier: str) -> bool:
        return identifier in self.variables or bool(
            self.parent and self.parent.has_variable(identifier)
        )

    def list_variables(self) -> Set[str]:
        all_variables = set(self.variables)
        if self.parent:
            all_variables |= self.parent.list_variables()
        return all_variables

    def reference_variable(
        self,
        identifier: str,
        node: AstNode,
        scope: Optional["LexicalScope"] = None,
    ):
        if not scope:
            scope = self

        if variable := self.variables.get(identifier):
            binding = variable.bindings[-1]

            if not binding.references or binding.references[-1] is not node:
                self.dirty = True
                binding.references.append(node)

                if variable.storage != "local" and self.parent:
                    self.parent.reference_variable(identifier, node, scope)

            return

        if not self.parent:
            raise set_location(UndefinedIdentifier(identifier, scope), node)

        self.parent.reference_variable(identifier, node, scope)

    def bind_variable(self, identifier: str, node: AstNode):
        self.dirty = True

        if self.has_binding(identifier):
            self.reference_variable(identifier, node)

        variable = self.variables.get(identifier)
        if not variable:
            variable = Variable()
            self.variables[identifier] = variable

        variable.bindings.append(Binding(origin=node))

        if variable.storage != "local" and self.parent:
            self.parent.bind_variable(identifier, node)

    def bind_storage(
        self,
        identifier: str,
        storage: Literal["nonlocal", "global", "deleted"],
        node: AstNode,
    ):
        self.dirty = True

        self.reference_variable(identifier, node)

        variable = self.variables.get(identifier)
        if not variable:
            variable = Variable(storage=storage)
            self.variables[identifier] = variable
        elif storage == "deleted":
            variable.storage = storage
        elif variable.storage != storage:
            exc = InconsistentIdentifierStorage(identifier, storage, self)
            raise set_location(exc, node)

        variable.bindings.append(Binding(origin=node))

    def push(self, scope_type: Type["LexicalScopeType"]) -> "LexicalScopeType":
        child = scope_type(parent=self)
        self.children.append(child)
        return child

    def deferred(self, scope_type: Type["LexicalScopeType"]) -> "LexicalScopeType":
        if not isinstance(self.next_deferred, scope_type):
            self.next_deferred = self.push(scope_type)
        return self.next_deferred  # type: ignore

    def deferred_complete(self):
        self.next_deferred = None

    def fork(self) -> "LexicalScope":
        return replace(
            self,
            parent=self.parent and self.parent.fork(),
            dirty=False,
            variables=self.fork_variables(),
        )

    def fork_variables(self) -> Dict[str, Variable]:
        return {
            identifier: variable.fork()
            for identifier, variable in self.variables.items()
        }

    def reconcile(self, lexical_scope: "LexicalScope"):
        if lexical_scope.dirty:
            self.dirty = True
            self.reconcile_variables(lexical_scope.variables)
            if self.parent and lexical_scope.parent:
                self.parent.reconcile(lexical_scope.parent)

    def reconcile_variables(self, variables: Dict[str, Variable]):
        for identifier, variable in variables.items():
            if current_variable := self.variables.get(identifier):
                current_variable.reconcile(variable)
            else:
                self.variables[identifier] = variable

    def create_pending_binding(self, identifier: str, node: AstNode):
        self.pending_bindings.append((identifier, node))

    def flush_pending_bindings(self):
        for identifier, node in self.pending_bindings:
            self.bind_variable(identifier, node)
        self.pending_bindings.clear()


@dataclass
class GlobalScope(LexicalScope):
    """Specialized global scope to fork more efficiently."""

    identifiers: Set[str] = field(default_factory=set)

    def has_variable(self, identifier: str) -> bool:
        return identifier in self.identifiers or super().has_variable(identifier)

    def list_variables(self) -> Set[str]:
        return self.identifiers | super().list_variables()

    def reference_variable(
        self,
        identifier: str,
        node: AstNode,
        scope: Optional["LexicalScope"] = None,
    ):
        if identifier not in self.identifiers or identifier in self.variables:
            super().reference_variable(identifier, node, scope)


class FunctionScope(LexicalScope):
    """Dedicated type for function scope."""


class MacroScope(FunctionScope):
    """Dedicated type for macro scope."""


class ProcMacroScope(FunctionScope):
    """Dedicated type for proc macro scope."""


class ClassScope(LexicalScope):
    """Dedicated type for class scope."""

    def push(self, scope_type: Type["LexicalScopeType"]) -> "LexicalScopeType":
        if issubclass(scope_type, FunctionScope):
            if not self.parent:
                raise ValueError("Class scope has no parent.")
            return self.parent.push(scope_type)
        return super().push(scope_type)
